# caserver-proxy

Proxies client requests to either a *sync* or *async* [caserver](https://gitlab.elettra.eu/puma/server/caserver)
instance, or both, depending on the client request.

Each client interested in monitoring *sources* through *nchan channels* (*"subscribe method"*) shall request a *client id* before 
subscribing and keep it valid on the server side by periodical *keep alive* messages. *client id* registration and renewal are 
performed on a dedicated URL.

The *caserver-proxy* registers clients on a database and as soon as they don't renew their subscription (they vanish) their 
sources are *unsubscribed* from the *caserver async instance*, that shall no longer monitor them. 

Don't forget to read the [FAQs](#faqs) below.

## Concepts

#### Request

Requests are HTTP requests issued by a client to the *ca-proxy*. Requests are proxied by *ca-proxy* and forwarded to the 
*caservers* and *ca-supervisor*.
Each request has a *method* specifying the desired operation (*read*, *s* (subscribe: read once and then monitor a source 
over time), *write*, *S* (subscribe only), *edit* (modify parameters of a monitored source) and so on).

#### Reply

Every HTTP request receives a *synchronous* reply. Depending on the request, the reply provided by the *ca-proxy* may either 
simply be the one from one *caserver* forwarded back (method: *read*) or the combination (merge) of the 
*async and sync caservers*' (method *s*).

#### Monitor over a channel

Subscribe requests (*s* or *S*) start a monitor on the specified source of data. The *async* caserver schedules the monitoring and 
replies immediately, generally with a simple *ok* message, if the JSON request is well formed. No data nor any other relevant 
information shall be expected in the reply to an *s* or *S* request.
Updates of the monitored data are published on the channel specified in the request ("channel":"chan_name" key/value).
 
## Architecture

Since the tasks of monitoring sources over time is radically different from a single synchronous *read* or *write* operation
from/to the control system engine, two instances of *caserver* shall run in parallel, each one with a dedicated task.
This shall keep each implementation as small and simple as possible, making debugging and resource management much easier.

### <a href="rationale" /> Rationale

##### 1. Each component of the architecture the simplest possible

We want to keep the *caserver* as simple and minimalist as possible. *caserver* shall be unaware of the supervisor and its 
interaction with the database shall be negligible. Only the essential task of recording a *heartbeat* timestamp shall be
carried out by the elementary [ca-supervisor-plugin](https://gitlab.elettra.eu/puma/server/caserver-supervisor-plugin).

##### 2. Stateless ca-proxy

The *ca-proxy* shall not rely on any state at any moment. It can be restarted at any time *transparently* from any other
architecture component's point of view.

#### Why a proxy? 

In principle, clients may issue individual requests for *read* and *subscribe* operations. This architecture would not require a proxy. Nevertheless, in harmony with the principles expressed in the [rationale](#rationale) :

- unsubscribes must be sent to all caserver instances from a centralized entity. Indeed, NGINX load balancer distributes 
subscribes across the available *caserver async* services: when srcX is unsubscribed we don't know which *caserver async* is in charge of it;

- subscribes and unsubscribes shall be forwarded to the [ca-supervisor](https://gitlab.elettra.eu/puma/server/ca-supervisor/-/blob/master/README.md) in order to be recorded on a database for potential recovery;

- we cannot force a client to gracefully unsubscribe on exit. Moreover, a client may stop unexpectedly unable to transmit the 
  unsubscribe request. On the other hand, it is required to send a *keepalive* message with an *ID* the client itself
  shall request before any subscription. The client IDs must be managed by a centralized entity (*ca-proxy*) and saved 
  into the database.
  A dedicated location *bu/tok* is reserved for *ID* management.

- in most cases graphical interfaces need source properties: minimum, maximum, alarm values, measurement unit and so on. A single 
  call to *caserver proxy* instead of two separate ones to different services simplifies the client side interface.

### Load balancing and failover

The integration with the [NGINX Load Balancer, Web Server, & Reverse Proxy](https://www.nginx.com/), 
the [ca-supervisor](https://gitlab.elettra.eu/puma/server/ca-supervisor) and the 
[caserver-supervisor-plugin](https://gitlab.elettra.eu/puma/server/caserver-supervisor-plugin) offers 
load balancing and failover mechanisms to the entire architecture.

### caserver sync

The *sync* instance of caserver shall *read* from or *write* to the control system engine and provide an immediate synchronous
reply to the client's request.

### <a href="caserver-async" /> caserver async

The *async* instance of caserver shall monitor sources over time, until they are explicitly *unsubscribed* (by either 
the client or the *ca-proxy* upon client expiration time). Upon receiving a request from the *ca-proxy*, it will also
give an immediate reply to the *ca-proxy* with either a single *ok* or multiple *error* messages. Nevertheless, only 
early errors are detected and notified (malformed requests, *async caserver* not responding), because source monitoring 
is just scheduled upon subscription and will start later. Errors from the engine shall be notified through the channel 
subsequently.

Depending on the client request, it will be forwarded by the *caserver proxy* to one or both *caservers*, merging 
the response from either (if necessary) into the HTTP reply.
If a network error occurs in the communication to one or both *caservers*, a 502 status reply shall contain the error
message(s) in the HTTP payload. If NGINX is configured with multiple upstream *ca-proxyies* for failover, the request 
will be routed to another available instance.


### <a href="recommendations" />  Requests: recommendations

Even though clients can pack requests together to perform different operations (called *methods*: "s", "read", "u", ...) 
within the same *HTTP request* to the *ca-proxy*, it is a good practise to send them grouped by *method*.
This improves performance on the server side (merging Json results from etherogeneous operations is unnecessary) and simplifies
the interpretation of the Json response for the client. With this observation in mind, a dedicated paragraph will discuss *mixed
method requests*.

### <a href="conf_opts" />  Requests: configuration options

The JSon *array* value associated to the *options* key within each *src* contains the *option* names that the client wants to set
on the cumbia engine. Within the same *src*, each option defined in the *options* array shall be added as a *key* associated to the
desired value.

This approach (option key list first defined and then option values assigned) favors flexibility at the price of a longer JSon request.
Options depend on the engine. They can be either added or removed and the adopted syntax decouples the client from the engine.

The most common options are the following:

##### value-only

- "value-only":"true"  in the context of either *s* or *conf* method, the synchronous reply shall bear only the read value, without the
  configuration associated to the source (in Tango, this means no database access for either attribute properties or command info);

- "value-only":"false" [ default ] in the context of either *s* or *conf* method, the synchronous reply shall carry the 
  *source configuration* alongside the source current *value*.

Note: this option does not apply to the *read* method, which always reads the current source value and never its configuration

The method *S* (capital *s*) can be used as an alternative to *s* to skip the *synchronous read*. 
This implies avoiding the configuration phase yet involves a potentially delayed first update for the client (depending on the caserver
*async* configuration).

##### no-value

- "no-value":"true"  in the context of either *s* or *conf* method, do not include the current *value* of the given *src*

- "no-value":"false" [ default ] the reply contains the value of *src*. This is the default behaviour in *conf* and *s* methods.

This option does not make sense within the *read* method.



##### read-value

- "read-value":"true"  *suggest* the engine that it is *always safe* to perform a reading contextually to a *subscribe* or *configuration*.

- "read-value":"false" [ default ] use a safer approach (needed for Tango commands) that does not perform a *read* if the *src* is potentially a *command* (Tango).
   The service uses the *CuTWriterConfFactory* if "read-value" is missing or set to false. If a command is detected (a/b/c->cmd), it is never
   imparted by the *synchronous* service at configuration time, because that would imply the command execution, which is unwanted.

When a client is sure that a (*Tango*) *command* can be *read* safely (without actually imparting any command), it shall set "read-value":"true" to get a 
read value included in the synchronous reply. For example, the Tango *State* command in the context of a *reader*.

##### period 

- "period":"INT" an integer value suggesting the polling period desired for the given *src*.

This is just a hint for the engine: according to the *async service* configuration, the polling period may be longer than required (lower frequency).
Remember also that either quick or *big sized* arrays (where again, *big* depends on the service configuration) may switch communication over *Websocket*
channels in *binary mode* (server version >= 1.5.0)


### <a href="subscribe" /> Subscribe requests, method: "s"

                          +----------+    s
    | client | --- s ---> | ca-proxy | ------> | caserver-sync |
          ^               |          |                |  <- reply from caserver sync
          |               |         /| <--------------+												
          +---------------|---<----+ |
            merged reply  |         \| <--------------+
                          +----------+                |
                                     |                |  <- reply from caserver async
                                     |    s           |
                                     | ------> | caserver-async|

  
__Diagram 1.__ Subscribe requests ("s" lowercase) are duplicated: a copy is sent to *caserver async* for monitoring, the 
other to *caserver sync* to get a synchronous readback. Replies are merged in the HTTP response to the client.
If either *caserver* fails, a 502 HTTP status is returned, to tell NGINX to forward the request to the next upstream.
    
A subscribe request identified by a *method* equal to lower case *s* for a given source of data (*src*),  
is issued by a client when a synchronous readback of the current value and further asynchronous updates 
on an *nchan channel* are required.

Subscribe requests from clients (*method: s*) imply

1. an immediate synchronous reply containing the value of the source 
2. a subscription for further updates through an *nchan* channel on which the client shall listen

Requirements 1. and 2. imply that a subscribe request shall be forwarded both to the *sync* and *async caserver instances*.
The replies from the *caservers* are then merged together by *caserver proxy* and sent back to the client as an http reply
to the client's http request.


The client shall expect in the reply a JSON an arrays containing two elements:

1. one array containing an object for each requested sources
2. one array containing an *ok* message or some elements describing the error of the *subscribe* operation on the *async caserver*.

Each error description shall contain at least:

- an "src" value, storing the source name
- an "err" field set to *true*
- an "msg" string with the error description

#### Example 1

The client sends an http request to *ca-proxy* in order to read two sources of data, *srv-tango-ctrl-01:20000/test/device/1/double_scalar*
and *srv-tango-ctrl-01:20000/test/device/1/long_scalar*, with the following payload in Json format:

```json
 {
     "channel":"qumbia-reader-ebe419b447235fec9b5f811c3468eb8c",
     "id":"223",
     "srcs":[{"method":"s","options":["period"],"period":"1000","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"},
     {"method":"s","options":["period"],"period":"1000","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]
 }
```

The expected reply is something like this:

```json
[{"_class_":0,"abs_change":"0.01","activity":"property","archive_abs_change":"0.1","archive_period":"1000","archive_rel_change":"0.1","argins":[],"color":"dark_green","data":true,"delta_t":"Not specified","delta_val":"Not specified","description":"Un doppio scalare del device primo","device":"srv-tango-ctrl-01:20000/test/device/1","df":0,"dfs":"scalar","dim_x":1,"disp_level":0,"display_unit":"pq","dt":5,"err":false,"format":"%4.2f","is_command":false,"label":"Doppio Scalare3","max":"255","max_alarm":"Not specified","max_dim_x":1,"max_dim_y":0,"max_warning":"Not specified","method":"s","min":"-255","min_alarm":"Not specified","min_warning":"Not specified","msg":"CuTConfigActivity.execute (1): ","name":"double_scalar","options":["period"],"period":"1000","periodic_period":"1000","point":"double_scalar","properties":[],"q":2,"qc":"white","qs":"valid","rel_change":"Not specified","root_attr_name":"Not specified","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar","standard_unit":"1.0","timestamp_ms":1642156353013.0,"timestamp_us":1642156353.01394,"type":"property","unit":"Pippoer","value":-66.25893320791678,"vt":10,"w_value":0.0,"writable":3,"writable_attr_name":"double_scalar"},{"_class_":0,"abs_change":"1","activity":"property","archive_abs_change":"1","archive_period":"2000","archive_rel_change":"1","argins":[],"color":"dark_green","data":true,"delta_t":"Not specified","delta_val":"Not specified","description":"No description","device":"srv-tango-ctrl-01:20000/test/device/1","df":0,"dfs":"scalar","dim_x":1,"disp_level":0,"display_unit":"km","dt":3,"err":false,"format":"%6d","is_command":false,"label":"long_scalar","max":"Not specified","max_alarm":"Not specified","max_dim_x":1,"max_dim_y":0,"max_warning":"Not specified","method":"s","min":"Not specified","min_alarm":"Not specified","min_warning":"Not specified","msg":"CuTConfigActivity.execute (1): ","name":"long_scalar","options":["period"],"period":"1000","periodic_period":"1000","point":"long_scalar","properties":[],"q":2,"qc":"white","qs":"valid","rel_change":"1","root_attr_name":"Not specified","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar","standard_unit":"1.0","timestamp_ms":1642156353013.0,"timestamp_us":1642156353.013942,"type":"property","unit":"tapiocasUnit","value":237,"vt":5,"w_value":0,"writable":3,"writable_attr_name":"long_scalar"},{"count":2,"id":284,"msg":"ok","type":"async"}]
```

The first two elements of the JSON array carry the data from *caserver sync*, the last one, with the key *type* set to *async* 
is the reply from the *async* instance, simply indicating that monitoring has been scheduled (see also 
[caserver async](#caserver-async)).

##### value-only

If the JSon request *options* key/value is modified as follows:

```json
"options":["period","value-only"],"period":"1000","value-only":"true",...
```

then the *src* configuration (min,max,label,format, ...) is not acquired.


##### no-value

If the JSon request *options* contains the *no-value* option and it is set to *true*, then the *caserver sync* will not perform a reading on the given src.
If the request is a *s* (subscribe), this may imply a longer interval for the client to receive the first value from the *async* service (from the channel).

If the configuration is only what you want, consider using the method *conf* instead.


##### <a href="ex_curl_webcon" /> Example 1 (curl + web console)

First we need to register the new client for the subscription. The *ca-proxy* associates an ID to each new client, so that if the latter vanishes without properly unsubscribing, the associated sources being monitored can be stopped.
A *time to live* parameter is configured in order to assert the maximum delay after which unused resources are stopped. 

The ID is obtained by getting:

> curl http://woody.elettra.eu:8001/bu/tok

The reply will be something like this:

```json
[{"exp":1641905106,"id":242,"ttl":300}]
```

We can see that the *id* is 242, the time to live is 300 seconds and the expiration UNIX timestamp is set to 300 seconds from now.
Within the time to live interval another call to the same URL shall be done in order to keep the client alive on the server side.

If we issue a second *curl* within 300 seconds from the first one, specifying *242* in the *id* field:

> curl http://woody.elettra.eu:8001/bu/src-bundle -d  $'{"channel":"rchan","id":"242","srcs":[{"method":"s","options":["period"],"period":"1000","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"},{"method":"s","options":["period"],"period":"1000","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]}'

we will get the same reply seen above. 

If we had previously *tuned* a web browser console on the channel *rchan* ("channel":"rchan" in the request above) by:

1. pointing a web browser page to http://woody.elettra.eu:8001/
2. Hitting Ctrl+Shift+I to open the web console (Firefox or Google Chrome)
3. Typing

 ```js
var source = new EventSource("http://woody.elettra.eu:8001/sub/rchan");
source.onmessage = onNewUpdate;
function onNewUpdate(message){
    console.log(message.data);
}
```
upon issuing the aforementioned curl request we would receive the events from the *rchan* channel until the client subscription expires.

If you issue the request without first registering the client, you receive a response like this:

```json
[{"err":true,"msg":"invalid id 241"}]
```

##### Example 2 (curl + web console)

Here's another example using curl and web console

1. Open a browser and point to https://puma-01.elettra.eu:8443/
2. Open the web developer console and type:

```
var srchan2 = new EventSource("https://puma-01.elettra.eu:8443/sub/mychan");
srchan2.onmessage = onNewUpdate;
function onNewUpdate(message){
    console.log("mychan ----->  " + message.data);
}
```

Please note that *mychan* is the name you choose for the channel to listen to.


3. On a terminal, request the *id* for the client:

> curl https://puma-01.elettra.eu:8443/bu/tok

```
[{"exp":1650551657,"id":16,"ttl":100}]
```

Note the expiration timestamp in case you want to keep alive the updates on the channel.

Subscribe to the channel (named *mychan*) using the new ID (number 16)

>  curl https://puma-01.elettra.eu:8443/bu/src-bundle -d  $'{"id":"16", "channel":"mychan",  "srcs":[{"chan":"mychan", "method":"s","options":["p"],"src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}'

You will receive the configuration data and the first read value *synchronously* on the console:

```
[{"_class_":2,"abs_change":"0.01","activity":"property","archive_abs_change":"0.1","archive_period":"1000","archive_rel_change":"0.1","argins":[],"color":"dark_green","data":true,"delta_t":"Not specified","delta_val":"Not specified","description":"Un doppio scalare del device primo","device":"srv-tango-ctrl-01:20000/test/device/1","df":0,"dfs":"scalar","dim_x":1,"disp_level":0,"display_unit":"DS","dt":5,"err":false,"format":"%4.2f","is_command":false,"label":"Doppio Scalare","max":"255","max_alarm":"225","max_dim_x":1,"max_dim_y":0,"max_warning":"100","method":"s","min":"-255","min_alarm":"-120","min_warning":"-100","msg":"CuTConfigActivity.execute (1): ","name":"double_scalar","options":["p"],"periodic_period":"1000","point":"double_scalar","properties":[],"q":2,"qc":"white","qs":"valid","rel_change":"Not specified","root_attr_name":"Not specified","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar","standard_unit":"1.0","timestamp_ms":1650551565997.0,"timestamp_us":1650551565.997515,"type":"property","unit":"Doubles","value":64.27891907267271,"vt":10,"w_value":100.0,"writable":3,"writable_attr_name":"double_scalar"},{"count":1,"id":16,"msg":"ok","type":"async"}]
```

And will start receiving updates on the browser console:

```
mychan ----->  [{"E":"change","dfs":"scalar","err":false,"mode":"E","msg":"","q":2,"qc":"white","qs":"valid","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar","timestamp_ms":1650551404997.0,"timestamp_us":1650551404.997569,"value":87.46192711632784,"vt":10,"w_value":100.0}] debugger eval code:4:13
mychan ----->  [{"E":"change","dfs":"scalar","err":false,"mode":"E","msg":"","q":2,"qc":"white","qs":"valid","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar","timestamp_ms":1650551406997.0,"timestamp_us":1650551406.997542,"value":88.2947163754854,"vt":10,"w_value":100.0}]
```

To stop the subscription:

> curl https://puma-01.elettra.eu:8443/bu/tok -d  $'{"id":"16", "method":"u"}'

You will see a "merged" reply to the *unsubscribe request* from all the *caserver async* instances, since 
*unsubscribe* requests are broadcasted to every *async* service:

```
[{"async-srv":5,"msg":"ok","replies":[[{"count":1,"msg":"ok","type":"async"}],[{"count":1,"msg":"ok","type":"async"}],[{"count":1,"msg":"ok","type":"async"}],[{"count":1,"msg":"ok","type":"async"}],[{"count":1,"msg":"ok","type":"async"}]]}] 
```

If not interested in the first configuration data, replace *s* with  capital *S* in the request:

> curl https://puma-01.elettra.eu:8443/bu/src-bundle -d  $'{"id":"21", "channel":"mychan",  "srcs":[{"chan":"mychan", "method":"S","options":["p"],"src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}'

Expected reply:

```
[{"count":1,"id":21,"msg":"ok","type":"async"}]
```

##### Example 3 (cumbia read)

Parameters: 
- -m: "monitor": first read is synchronous (caserver sync), then monitor over time until a key is hit.
- -c rchan: give a name to the channel (that you can use on the web console to tune to the channel)
- -u *ca-proxy* URL 
- list of sources (in the example: Tango engine)

> cumbia read -m -c rchan -u http://woody.elettra.eu:8001 test/device/1/double_scalar test/device/1/long_scalar

Given the nature of Server Sent Events on the *rchan* nchan channel, tuning a browser as suggested in the former paragraph
will show the updates on the web console as well.

The *cumbia read* command line utility gracefully unsubscribes when exiting, so that the *async caserver* stops monitoring 
immediately (the web console immediately stops receiving messages from *rchan*).


### <a href="read" /> Read requests, method "read"

                             +----------+  read
    | client | --- read ---> | ca-proxy | ------> | caserver-sync |
          ^                  |          |                |  
          |                  |          |                |  <- reply from caserver sync							
          +----------------- |          | <--------------+
             read reply      +----------+ 
                                                                      

                                                   | caserver-async|   <-- __uninvolved!__

__Diagram 2.__ Read requests are transparently proxied to *caserver-sync* only
  
Read requests ("method":"read") are forwarded only to the *caserver sync* instance because they require a single 
reading and an immediate synchronous reply. No monitoring or channels involved. 
Nevertheless, the client shall expect a reply enclosed in a top level *one element* array. As an example, if the client
requires the value of two sources, a two element array with the respective results shall be enclosed into a one element 
super array. Please note that in this case neither *id* nor *channel* are involved; the method is *read*:

> curl http://woody.elettra.eu:8001/bu/src-bundle -d  $'{"srcs":[{"method":"read","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"},{"method":"read","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]}'

```json
[
    [{"_class_":0,"activity":"property","argins":[],"color":"dark_green","data":true,"device":"srv-tango-ctrl-01:20000/test/device/1","dfs":"scalar","err":false,"is_command":false,"method":"read","msg":"CuTConfigActivity.execute (1): ","point":"double_scalar","properties":[],"q":2,"qc":"white","qs":"valid","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar","timestamp_ms":1641907208404.0,"timestamp_us":1641907208.404769,"type":"property","value":116.35477612972747,"value-only":true,"vt":10,"w_value":240.0},{"_class_":0,"activity":"property","argins":[],"color":"dark_green","data":true,"device":"srv-tango-ctrl-01:20000/test/device/1","dfs":"scalar","err":false,"is_command":false,"method":"read","msg":"CuTConfigActivity.execute (1): ","point":"long_scalar","properties":[],"q":2,"qc":"white","qs":"valid","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar","timestamp_ms":1641907208404.0,"timestamp_us":1641907208.404771,"type":"property","value":190,"value-only":true,"vt":5,"w_value":0}
    ]
]
```

Another observation: the reply obtained here is exactly the same as the one obtained with the *s* method. Indeed the lower case 
*s* method involves a first synchronous read, as already discussed. In the reply above, the outer array has been put in evidence.
The inner array contains the two read results, one for *double_scalar*, on for *long_scalar*.

#### <a href="read_note1" />Important note

The *Tango* engine *does not get database properties* upon "read".
The *property* option shall be explicitly set to "true" to fetch database properties alongside the read value:

> curl http://taeyang.elettra.eu:8001/bu/src-bundle -d  $'{"srcs":[{"method":"read","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar", "options":["property"],"property":"true" }]}'


### unsubscribe requests, method:"u"

Unsubscribe requests are forwarded by *ca-proxy* to the *caserver async* instance only.
Actually, in an architecture where *multiple ca-proxy / caserver* instances are in use for load balancing and 
failover, the initial subscribe request (*s*) was sent to a *caserver* by the *NGINX* proxy according to the 
balancing policy and system load at the time. When unsubscribing, we don't know which *caserver async* is in 
charge of the source involved.

The *ca-proxy will notify all caservers of an unsubscribe request*. The *caserver* in charge of that source will stop
monitoring it and release the associated resources.


                          +----------+    u
    | client | --- u ---> | ca-proxy | ------> | *caserver-async-1* |
          ^               |          |                | 
          |               |         /| <--------------+												
          +---------------|---<----+-| <--------------------------------+
            merged reply  |         \| <--------------+                 |
                          +----------+                |                 ^
                                     |                |                 |
                                     |    u           |                 |
                                     | ---> | *caserver-async-2* |      |
                                     |                                  |
                                     |                                  |
                                     .                                  .
                                     .                                  .
                                     .                                  .
                                     |     u                            |
                                     +-----> | *caserver-async-n* |->---+
                                                         


__Diagram 3.__ The unsubscribe request *u* from the client is *replicated n 
times* and *broadcasted* to all instances of caserver-sync. The HTTP response to the client
carries the merged replies from the N *caservers*                                                

There are a few ways to unsubscribe one or more sources. Each of them require the client ID.
Let's discuss them with examples.

#### Method 1: globally, by client id.

The following *curl*

> curl http://woody.elettra.eu:8001/bu/tok -d  $'{"id":"250", "method":"u" }'

will unsubscribe *all sources linked to the client with id 250*.

#### Method 2: by client id and source (or sources)

> curl http://woody.elettra.eu:8001/bu/src-bundle -d  $'{"id":"249", "srcs":[{"method":"u","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]}'

will unsubscribe *srv-tango-ctrl-01:20000/test/device/1/long_scalar* for the client with id 249.

As discussed in the paragraph below, if the same source is monitored by some other caserver or even by the same caserver for 
different clients, the service shall guarantee that the latter will be unaffected. 

#### Method 3: by client id, source and channel 

Sources can be subscribed also in relation to their channel.
We proceed with an example. 
First, get a new client ID:

```bash
curl http://woody.elettra.eu:8001/bu/tok
```

A web console will monitor data over the channels.
The following subscribe request monitors "double_scalar" on channel "rchan" and "long_scalar" on "chan1", in the hypotesis
that the above call returned an id equal to *303*:

```bash
curl  http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"303", "srcs":[{"channel":"rchan", "method":"s","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"},{"channel":"chan1", "method":"s","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]}'
```

__Note.__ If you need more time to carry out the experiment, don't forget to renew the client ID obtained with the first *curl* 
call before the expiration time (*"ttl"*) indicated in that response:
> curl http://woody.elettra.eu:8001/bu/tok -d  $'{"id":"303" }'



You will need this additional code on the web browser console in order to receive events on *chan1* alongside *rchan* 
(refer to [this example](#ex_curl_webcon)):

```js
var source = new EventSource("http://woody.elettra.eu:8001/sub/chan1");
source.onmessage = onNewUpdate;
function onNewUpdate(message){
    console.log(message.data);
}
```

The curl command

```bash 
    curl  http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"303", "srcs":[{"channel":"rchan", "method":"u","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}'
```

will unsubscribe "double_scalar" on channel "rchan" (check the web console!)

While the following:

```bash
 curl  http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"303", "srcs":[{"channel":"chan1", "method":"u","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]}'
```

will stop "long_scalar" for the client 303 on channel "chan1".
In this case, specifying the channel is redundant because the two sources are distinct.

Now consider this case, where "double_scalar" is monitored twice on "ch1" and "ch2"

```bash
curl  http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"305", "srcs":[{"channel":"ch1", "method":"s","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"},{"channel":"ch2", "method":"s","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}'
```

Adjust the monitor on the web console to tune to *ch1*:

```js
var srchan = new EventSource("http://woody.elettra.eu:8001/sub/ch1");
srchan.onmessage = onNewUpdate;
function onNewUpdate(message){
    console.log("chan 1 ---> " + message.data);
}
```

and similarly to *ch2*.

The command 

```bash
curl  http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"305", "srcs":[{"channel":"ch1", "method":"u","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}'
```

will stop "double_scalar" on *ch1* only.

You need to issue

```bash
curl  http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"305", "srcs":[{"channel":"ch2", "method":"u","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}'
```

to stop monitoring "double_scalar" on *ch2*.

#### Expected reply

Since unsubscribe requests are broadcasted to all *caserver async* servers listed in the *ca-proxy* configuration file, the 
client shall expect a reply carrying a payload with a JSON array each element being a message from one *caserver*. In this 
example, two *caserver async* instances were specified in the configuration file though at the moment one was offline:

```json
[{"async-srv":2,"err":true,"msg":"ok","replies":[[{"count":2,"id":285,"msg":"ok","type":"async"}],{"err":true,"msg":"192.168.1.159:19294: Connection refused"}]}]
```

This second example shows a reply from two *caserver async* instances both online and acknowledging the unsubscribe request:

```json
[{"async-srv":2,"msg":"ok","replies":[[{"count":1,"id":0,"msg":"ok","type":"async"}],[{"count":1,"id":0,"msg":"ok","type":"async"}]]}]
```


#### Observations on methods 1 and 2.

1. *Method 1 shall always be preferred by clients when they shutdown*

2. Method 1 is used by *ca-proxy* when a client *vanishes* (i.e. it has stopped sending the *keep alive* message).
   In fact, if a client exits without gracefully unsubscribe, the *ca-proxy* must take care of stray sources still 
   being monitored uselessly.

3. Method 2 is used by clients when they want to unsubscribe a subset of sources at a certain point at runtime.

#### Important note for client design.

The same source could be monitored by two or more different *caserver async* instances at the same time.
That is because client1 subscribed srcX on channel c1 at time t1, and later at time t2 client2 subscribed the same srcX
on channel c2. At t1 NGINX assigned (s1,c1,clid1) to caserver1, and (s2,c2,clid2) to caserver2.

         |
         |                           +-------+
         | s                         |       |  s
    cli1 -----> {srcX,c1,clid1} ---> | NGINX |-------->  | caserver1 |
         |                           |       |
         | s                         |       |  s
    cli2 -----> {srcX,c2,clid2} ---> |       |----------------->  | caserver2 |
         |                           |       |
         |                           +-------+
         |
         | time   
         v 

         
When cli2 unsubscribes srcX, clid2 shall be specified, so that caserver2 shall release srcX while caserver1 still 
serves srcX to cli1.

In the situation in which *NGINX* assigns {srcX,c1,clid1} and {srcX,c2,clid2} to the same *caserver async* instance,
for example caserver1:


         |
         |                           +-------+
         | s                         |       |  s
    cli1 -----> {srcX,c1,clid1} ---> | NGINX |-------->  | caserver1 |
         |                           |       |                 ^
         | s                         |       |  s              |
    cli2 -----> {srcX,c2,clid2} ---> |       |-----------------+  
         |                           |       |             
         |                           +-------+             | caserver2 |
         |
         | time   
         v 


caserver1 shall keep two references to srcX, one for clid2 and one for clid1, so that when cli2 unsubscribes, the reference 
count for srcX shall be *1* and still linked to {c1,clid1}.

##### Warning

The case when the same client *cli1* subscribes K times the same source srcX shall be avoided on the client side 
(more in general, by the client library). Even in the case the single *caserver* increases the reference for a duplicated source, if 
*NGINX* distributes srcX across different *caservers*, an unsubscribe request with *id clid1*, would be delivered to 
all *caservers*, stopping the updates on several or potentially all the K-1 handles on *cli1* still subscribed.

***The client [library] shall distribute the results from the link to srcX across all the K result listeners and
shall subscribe srcX to *ca-proxy* only once. (And unsubscribe only once no listeners need it)*** 
In other words, the client shall keep a reference count to handle duplicated sources on its own.

### Mixed method requests

It is possible to request operations with mixed methods in the same HTTP request.
The following one specifies a single *read* from *test/device/1/double_scalar* and a monitor on *test/device/1/long_scalar*.

> curl http://woody.elettra.eu:8001/bu/src-bundle -d  $'{"id":"251", "srcs":[{"method":"read","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"},{"channel":"rchan",  "method":"s","src":"srv-tango-ctrl-01:20000/test/device/1/long_scalar"}]}'

The response is always the same as the one already seen in the *subscribe* and very similar to the one discussed in the *read* paragraphs, because both *read* and *s* imply a synchronous response from *caserver sync*. In particular, this response is equal
to the one seen for the *subscribe* case. It contains as second element of the top level array the response from *caserver async*:

```json
[{"msg":"ok"}]
```

that, in this case, refers only to the sole source being subscribed (test/device/1/long_scalar*), even though this is not evident
from the *caserver async* "ok" message.


# <a href="proxy_bypass" /> Proxy bypassing for testing

### Testing caserver sync

*caserver sync* can be tested directly from clients bypassing the *ca-proxy*.
Insert the following section in the *nginx.conf* configuration file, where the port number
*8002* is used to bypass the proxy and send the requests to the *sync* instance listening 
on port number *9293*:

```
  server {
	listen 8002;
      server_name  woody.elettra.eu;
      location ~ /bu/(.+)$ {
          proxy_set_header Host $host;
          proxy_set_header X-Source $1;
          # send requests to caserver async listening on port 9293
	        proxy_pass http://taeyang.elettra.eu:9293;
       }
  
    } # server
```

In this case, clients connecting to the port 8002 will contact the *caserver async* directly.
As a consequence, if the client *subscribes* to a source for monitoring, an error shall be expected:

> cumbia read -m -u http://taeyang.elettra.eu:8002 test/device/1/double_scalar -c rchan
```
 * cumbia read: using default historical database profile, if available
 * monitor started: press any key to exit
!! error: CuHttpCliIdMan::onError: error: 
!! error: CumbiaHttp.onIdManError: client id manager error: Error transferring http://taeyang.elettra.eu:8002/bu/tok - server replied: Bad Request
 * double scalar: [14:43:54+0.953426]  [double] -79.11
```

You can see the value obtained by *caserver sync*: -79.11 and an error due to the fact that *cumbia read* 
sends a token request (with destination *caserver async* through the *ca-proxy*) to monitor the source.


# Security and system robustness

*NGINX* proxy shall be configured to protect the *caservers* from external flooding or any other security risk.
In other words, the *caserver* services *shall not be directly reachable from the Internet*. Although designed 
with security in mind, ill formed packets may hamper the service performance.

Using netcat we try to forge ill-formed packets and send them to the *ca-proxy*.

First, use a valid *HTTP request* to read *test/device/1/double_scalar*:

```
[giacomo@taeyang ~]$ printf $'POST /bu/src-bundle HTTP/1.0
Accept: application/json
Accept-Encoding: gzip, deflate
Accept-Language: en-US,*
Connection: close
Content-Length: 193
Content-Type: application/json
Host: taeyang.elettra.eu
User-Agent: cumbia-http 1.4.0
X-Source: src-bundle

{"srcs":[{"method":"read","options":["period","read-value","single-shot"],"period":"1000","read-value":"true","single-shot":"true","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}' | nc taeyang.elettra.eu 8001
```

We will get a valid reply:

```
HTTP/1.1 200 OK
Server: nginx/1.21.6
Date: Tue, 15 Feb 2022 10:16:56 GMT
Content-Type: application/json
Content-Length: 622
Connection: close
Accept: application/json
Accept-Encoding: gzip, deflate
Accept-Language: en-US,*
Host: taeyang.elettra.eu
User-Agent: cumbia-http 1.4.0
X-Source: src-bundle

[{"_class_":0,"activity":"property","argins":[],"color":"dark_green","data":true,"device":"srv-tango-ctrl-01:20000/test/device/1","dfs":"scalar","err":false,"is_command":false,"method":"read","msg":"CuTConfigActivity.execute (1): ","options":["period","read-value","single-shot"],"period":"1000","point":"double_scalar","properties":[],"q":8,"qc":"orange","qs":"warning","read-value":"true","single-shot":"true","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar","timestamp_ms":1644920130997.0,"timestamp_us":1644920130.997308,"type":"property","value":207.10863006520682,"value-only":true,"vt":10,"w_value":0.0}]
```

Now without altering the content length add extra characters after the json payload:

```
[giacomo@taeyang ~]$ printf $'POST /bu/src-bundle HTTP/1.0
Accept: application/json
Accept-Encoding: gzip, deflate
Accept-Language: en-US,*
Connection: close
Content-Length: 193
Content-Type: application/json
Host: taeyang.elettra.eu
User-Agent: cumbia-http 1.4.0
X-Source: src-bundle

{"srcs":[{"method":"read","options":["period","read-value","single-shot"],"period":"1000","read-value":"true","single-shot":"true","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}la reconch' | nc taeyang.elettra.eu 8001
```

We will still get a correct reply because the *HTTP parser* stops parsing the payload at Content-Length bytes.

Issuing the same *netcat* command above, this time altering the Content-Length to 196, we get the following reply,
with no consequences for neither the *ca-proxy* and the *caserver*:

```
HTTP/1.1 400 Bad Request
Server: nginx/1.21.6
Date: Tue, 15 Feb 2022 10:36:19 GMT
Content-Type: application/json
Content-Length: 36
Connection: close

[{"err":true,"msg":"no srcs found"}]
```

*ca-proxy* logs will evidentiate a JSON parse error. In this case, the *caserver* will not even receive the malformed packet.

Next time, we set a Content-Length of 1400 bytes and issue the same request. Since the request is first managed by the *NGINX* reverse proxy, this malformed request will not even be delivered to the *ca-proxy* and the connection will be closed after a while by the reverse proxy.


The *NGINX* reverse proxy protects also against *non http* content:

```
printf $'{"srcs":[{"method":"read","options":["period","read-value","single-shot"],"period":"1000","read-value":"true","single-shot":"true","src":"srv-tango-ctrl-01:20000/test/device/1/double_scalar"}]}la reconchita' | sudo nc -vD taeyang.elettra.eu 8001
```

```
Connection to taeyang.elettra.eu (192.168.205.25) 8001 port [tcp/vcom-tunnel] succeeded!
HTTP/1.1 400 Bad Request
Server: nginx/1.21.6
Date: Tue, 15 Feb 2022 10:51:52 GMT
Content-Type: text/html
Content-Length: 157
Connection: close

<html>
<head><title>400 Bad Request</title></head>
<body>
<center><h1>400 Bad Request</h1></center>
<hr><center>nginx/1.21.6</center>
</body>
</html>
```

## <a href="database">Database</a>

A Postgres database is used to store information about running services, monitor activities
and recovery operations in one centralized place.
The database is made up of several tables, each one is discussed in the documentation of 
the project that (mainly) accesses it in *write* mode. For example, *caserver-proxy* modifies 
the *register* table described below, while the 
[ca-supervisor](https://gitlab.elettra.eu/puma/server/ca-supervisor) mostly handles the *activity*
and recovery related tables.
The *service* and *srvconf* tables are both accessed in read-write mode by *ca-proxy* and the 
[caserver-supervisor-plugin](https://gitlab.elettra.eu/puma/server/caserver-supervisor-plugin).
They are described below in this document.


The *caserver-proxy* accesses the *register, service, clients and srvconf* tables.

#### register table

The *register* table 

```
CREATE TABLE register (
    client_id SERIAL PRIMARY KEY,
    time timestamp NOT NULL
);
```

records the "heartbeats" of each client with *client_id*. A *new client* is first automatically registered
by the *caserver-proxy*, that will release the *ID* assigned to the client. The latter shall refresh the 
record periodically thereafter. 
More details [here](#example-1-curl-web-console).

#### <a href="service-clients-srvconf-tables">service, clients and srvconf tables</a>

The three tables are quite simple:

```
CREATE TABLE srvconf (
    conf_id SERIAL UNIQUE,
    srvnam varchar(255),
    addr varchar(64) NOT NULL,
    port integer NOT NULL,
    PRIMARY KEY(addr,port)
);

CREATE TABLE service (
    id SERIAL PRIMARY KEY,
    started timestamp,
    expected timestamp,
    conf_id integer REFERENCES srvconf (conf_id)
);

CREATE TABLE "clients" (
    id integer PRIMARY KEY REFERENCES register (client_id) ON DELETE CASCADE,
    addr varchar(64) NOT NULL,
    FOREIGN KEY(id) REFERENCES register(client_id)
);

/* get table desc and their comments with \d+ */
COMMENT ON clients IS 'accessed by ca-proxy when a new client registers';
```

Please note that the [caserver-supervisor-plugin](https://gitlab.elettra.eu/puma/server/caserver-supervisor-plugin)
eiher adds a new row into the *service* table or updates an existing one when an instance of the 
*caserver async* enters the *listen* state. 

Also note that, to allow the system to work *without the caserver-supervisor-plugin*,
the *ca-proxy* adds a new row in order to return an *ID* for the client.
In fact, the *service.id* is a *SERIAL PRIMARY KEY* (ca-registrar.cpp). In this case, *ca-proxy* 
fills the new row with null values except for the *conf_id* that is identified by the pair
*async-server-address* and *async-server-port* taken from the configuration file.

## <a href="stress_tests">Stress tests</a>

Stress tests have initially involved only the *sync* instance of the caserver, which is deemed more critical since 
every request requires an access to the Tango control system and a synchronous reply.

Tests have been performed both in [proxy bypass mode](#proxy_bypass) and using the *ca-proxy* modified in order 
not to forward requests to the *async* server.

The [sequencer](https://gitlab.elettra.eu/cs/gui/sequencer) application, developed at Elettra, has been used to 
generate peaks of requests to the services. The applications creates a big tree of more than 1400 items and initial 
data updates can be more than 1900. During the tests, 24 to 40 instances of the application have been started 
at the same time. The application start up completes successfully.

#### Notes

1. In order to avoid *NGINX Gateway Timeout* errors during the stress test, the *proxy_read_timeout* has been set to 
   *300s* in the *location* section in *nginx.conf*.

2. Client *id* up to date check has been disabled in *CaRegistrar::m_record_up_to_date*, that always returned true.
   The sequencer app takes up a lot of resources during set up and also the client id update process on the server 
   may not keep up to the required rate. For the purpose of testing the engine, this choice can be acceptable. 
   *invalid id* error messages shall be expected otherwise.
   

# <a href="faq" />FAQ 

**Q.** The "read" method does not retrieve the Tango database properties.
**A.** This is the default implementation for *read* requests. See [this important note](#read_note1).

**Q.** I see *engine timeout* error messages received on the client side. Why?
**A.** A configuration option *caserver-sync:engine-timeo* determines how much the *caserver sync* shall wait 
       for a reply from the *Tango engine*. This may be due to the Tango control system, but *engine timeout*
       errors have been observed under heavy stress tests (sequencer stress test, 40 instances run concurrently).
       This happened especially after the *cumbia 1.4* [Tango device factory optimization](https://github.com/ELETTRA-SincrotroneTrieste/cumbia-libs/blob/master/cumbia-tango/CHANGES-1.4.md). A long queue of events accumulated
       on the thread turned out to require more than the default 10 seconds *engine timeout* between 
       *Cumbia::registerActivity* and *CuThread* processing the event in the queue.
       **Solution**: set a higher *engine timeout* on the configuration file.

