#include "ca-proxy-process.h"
#include <unistd.h>
#include <cahttpmsg.h>
#include "caproxyjson_h.h"
#include "caproxyhttph.h"
#include "caproxythreadpool.h"
#include "ca-registrar.h"
#include <calog.h>
#include <cajson-data-encoder.h>
#include <ca-receiver-a.h>
#include <cumbia.h>
#include <cuthreadfactoryimpl.h>
#include <cuthreadseventbridge.h>
#include <cutimer.h>
#include <cutimerservice.h>
#include <cuserviceprovider.h>
#include <cuservices.h>

class CaProxyPPrivate{
public:
    CaProxyPPrivate(const CaProxyThreadPool *p, const CaRegistrar *r, Cumbia* c)
        : propool(p), registrar(r), cumbia(c) {    }

    CuLog* log;
    std::string errmsg;
    const CaProxyThreadPool *propool;
    const CaRegistrar *registrar;
    Cumbia* cumbia;
    CuTimerService *timer_s;
    CuTimer *timer;
    CaProxyHttpH phttph;
    unsigned long long eventcnt;
};

CaProxyProcess::CaProxyProcess(Cumbia* cumbia,
                               const CuData &opts,
                               CuLog* log,
                               const std::string& servaddr,
                               const unsigned short servport,
                               const CaRegistrar *registrar,
                               const CaProxyThreadPool *thp) {
    d = new CaProxyPPrivate(thp, registrar, cumbia);
    d->log = log;
    // every expire-check-interval seconds, send the list of alive clients
    int interval = m_set_expcheck_interval(opts);
    d->timer_s = new CuTimerService();
    d->timer = d->timer_s->registerListener(this, static_cast<int>(interval * 1.05), static_cast<CuEventLoopService *>(cumbia->getServiceProvider()->get(CuServices::EventLoop)));
    d->eventcnt = 0;
}

CaProxyProcess::~CaProxyProcess() {
    d->timer_s->unregisterListener(this, d->timer->timeout());
    delete d->timer_s;
    delete d;
}

std::string CaProxyProcess::error() const {
    return d->errmsg;
}

void CaProxyProcess::onProgress(int , int , const CuData &) { }

// results from
// - receiver-a
// specific listeners for ca-proxy and ca-registrar
void CaProxyProcess::onResult(const CuData &data) {
    bool ok, unsub = false;
    const int& cli_fd = data.i("fd");
    if(data.containsKey("fd") && data.containsKey("fdopen")) {
        bool o =  data.B("fdopen");
        const int fd = data.I("fd");
        d->propool->fd_state_change(fd, o);
    }
    if(data.b("err")) {
        // error handling
        // - errors from ca-receiver-a
        d->log->write(data.s("from"), data.s("msg"));
    }
}

void CaProxyProcess::onResult(const std::vector<CuData> &) { }


void CaProxyProcess::onResult(const CuUserData *ud) {
    bool ok, unsub;
    if(ud->type() == CaReceiverData::CaReceiverDataType) {
        const CaReceiverData *rd = static_cast<const CaReceiverData *>(ud);
        //    CaNetMsg nmsg(data.s("data"), data.I("fd"), data.s("addr"), data.B("err"));
        const CaHttpMsg* msg = rd->msg;
        int cli_fd = rd->fd;
        if(msg->x_source == "tok") {
            // get token id. unsub will be true if "method":"u" is found in the json input
            std::string payload = std::string(msg->payload());
            unsigned long long id = CaProxyJsonH().tokmsg_process(payload,  &unsub, d->errmsg);
            if(d->errmsg.length() > 0) {
                d->propool->msg_send_cli(cli_fd, CaProxyEvent::CaPHttp400, d->phttph.proc_error(d->errmsg)); // error response; // error response
            }
            else  {
                // m_token_event sends event to ca-registrar, which can insert, update or delete (on unsubscribe) a record
                // the reply is delivered by caproxy thread pool/caproxy through caregistrar listener
                ok = m_token_event(id, /*data.s("addr")*/ "", rd->fd, unsub);
                if(!ok && cli_fd > 0) // !ok if database not connected
                    d->propool->msg_send_cli(cli_fd,  CaProxyEvent::CaPHttp500, d->phttph.proc_error("database not connected")); // error response
            }
        }
        else {
            ++d->eventcnt;
            // proxy operation
            ok = m_data_redistrib(cli_fd, msg);
            if(!ok && cli_fd > 0) { // !ok if json parse fails or empty json content in nmsg.payload
                d->propool->msg_send_cli(cli_fd, CaProxyEvent::CaPHttp400, d->phttph.proc_error(d->errmsg)); // error response
            }
        }
    }
    delete ud;
}

CuData CaProxyProcess::getToken() const {
    return CuData("ca-regi-process");
}

void CaProxyProcess::onTimeout(CuTimer *t) {
    d->timer_s->restart(t);
    d->registrar->new_event(new CaRegiEvent()); // check expired
}

/*!
 * \brief call do_proxy if the network message contains a valid request
 *        or return false in case of error.
 *
 *        The caller will send an error message back to the client without
 *        forwarding the request in case this method returns false
 * \param fd  client file descriptor
 * \param nmsg  message
 * \return true if proxy has been done, false otherwise
 */
bool CaProxyProcess::m_data_redistrib(int fd, const CaHttpMsg *nmsg) {
    std::string syr, asy_sub_r, asy_unsub_r;
    CaProxyJsonH jshelper;
    const std::string& injs = std::string(nmsg->payload());
    const std::string& hdr = std::string(nmsg->header());
    unsigned long long id;
    bool ok = jshelper.divide(injs, syr, asy_sub_r, asy_unsub_r, &id, d->errmsg, d->propool->srv_id());
    if(!ok)
        d->log->write("ca-regi-process", d->errmsg);
    else if(!(ok = (asy_sub_r.length() > 0 || syr.length() > 0 || asy_unsub_r.length() > 0) ) ) {
        d->errmsg = std::string("ca-regi-process: no valid requests") + ((id == 0) ? " missing id in root json?" : "");
    }
    else {
        // sync only operation requested: ok do proxy
        // async operations involved: do_proxy only if ID valid
        if( ((asy_sub_r.length() > 0 || asy_unsub_r.length() > 0) && id > 0 && (ok = d->registrar->id_valid(id))) ||
                (asy_sub_r.length() == 0 && asy_unsub_r.length() == 0 && syr.length())  > 0) {
            d->propool->do_proxy(fd, CaProxyEvent::CaPHttpPost, syr, asy_sub_r, asy_unsub_r, hdr, nmsg->clen);
        }
        else
            d->errmsg = "invalid id " + std::to_string(id);
    }
    return ok;
}

// process a POST /bu/tok request:
// if method is "u" proxy the unsubscribe event
// otherwise either get a new client id or renew an existing one through the registrar
//
bool CaProxyProcess::m_token_event(const unsigned long long &tok,
                                   const std::string &addr,
                                   const int fd,
                                   bool unsub) {
    bool ok = d->registrar->db_connected();
    if(ok)
        d->registrar->new_event(new CaRegiEvent(tok, addr, fd, unsub));
    return ok;
}

inline int CaProxyProcess::m_set_expcheck_interval(const CuData& opts) const {
    int clittl = opts.I("client-record-ttl") > 30 ? opts.I("client-record-ttl") : 30;
    int interval = opts.I("expire-check-interval");
    if(interval < clittl) {
        if(interval > 0)
            d->log->write("ca-proxy-process", CuLog::LevelError, "expire-check-interval [%d] cannot be less than client-record-ttl [%d]: setting expire-check-interval to %d", interval, clittl, static_cast<int>(clittl * 1.04));
        interval = clittl * 1.04;
    }
    d->log->write("ca-proxy-process", CuLog::LevelAll, "client-record-ttl: %d: expire-check-interval %d [seconds]", clittl, interval);
    return interval *= 1000; // ms
}
