#include "ca-registrar.h"
#include <cadbhandle.h>
#include <cudata.h>
#include <mutex>
#include <stdexcept>
#include <shared_mutex>
#include <condition_variable>
#include <queue>
#include <list>
#include <iostream>
#include <sstream>
#include <locale>
#include <iomanip>

#include <assert.h>

class CaRegistrarPrivate {
public:
    CaRegistrarPrivate(const CuData& cp) : connp(cp), err(false), quit(false), id(0) {
        cp.containsKey("client-record-ttl") ? cli_ttl = cp.I("client-record-ttl") : cli_ttl = 300;
    }

    CaDbH *dbh, *dbh2;
    const CuData connp;
    std::queue <CaRegiEvent *> eq;
    std::mutex mu;
    std::condition_variable wait;
    bool err, quit;
    std::string msg;
    unsigned long long id;
    time_t cli_ttl;
    long srv_id; // from database

    pthread_t this_thread;
};

CaRegistrar::CaRegistrar(const CuData& connp)
    : CuActivity(CuData("type", "ca-registrar")) {
    d = new CaRegistrarPrivate(connp);
    // psql connection handle to use outside the activity thread
    // psql is thread safe if we use one handle per thread
    d->dbh2 = new CaDbH(connp.s("dbuser"), connp.s("dbpass"), connp.s("dbnam"), connp.s("dbhost"), connp.s("dbport"));
    if(!d->dbh2->connect()) {
        publishResult(CuData("err", true).set("msg", d->dbh2->msg));
        d->dbh2 = nullptr;
    }
}

CaRegistrar::~CaRegistrar() {
    delete d;
}

void CaRegistrar::new_event(CaRegiEvent *e)  const{
    std::unique_lock<std::mutex> lock(d->mu);
    d->eq.push(e);
    d->wait.notify_one();
}

std::string &CaRegistrar::msg() const {
    return d->msg;
}

int CaRegistrar::clientRecordTtl() const {
    return d->cli_ttl;
}
void CaRegistrar::setClientRecordTtl(int sec) {
    d->cli_ttl = sec;
}

bool CaRegistrar::id_valid(const unsigned long long &id)  const{
    return  m_record_up_to_date(id, d->dbh2);
}

bool CaRegistrar::db_connected() const {
    return d->dbh->connected;
}

void CaRegistrar::stop() {
    d->quit = true;
    d->wait.notify_one();
}

// returns new ID
unsigned long long CaRegistrar::m_insert_record(const std::string &addr, time_t& expires) {
    assert(d->this_thread == pthread_self());
    expires = 0;
    std::string s_id = "0";
    std::string stmt = "INSERT INTO register (time) VALUES (current_timestamp) RETURNING client_id";
    std::vector<std::string> params;
    CaDbRes r = d->dbh->execute(stmt, params);
    if(!r.error() && r.size() == 1) {
        s_id = r.value(0, "client_id");
        stmt = "INSERT INTO clients (id, addr) VALUES ($1, $2)";
        params =  { s_id, addr };
        r = d->dbh->execute(stmt, params);
        time(&expires);
        expires += d->cli_ttl;
    }
    d->err = r.error();
    d->msg = r.errmsg;
    printf("CaRegistrar::m_insert_record error flag %d err msg %s id %s addr %s\n", d->err, d->msg.c_str(), s_id.c_str(), addr.c_str());

    return !d->err ? std::stoull(s_id) : 0ULL;
}

// return expires time
unsigned long long CaRegistrar::m_update_record(const unsigned long long &id, const std::string &addr, time_t& expires)
{
    assert(d->this_thread == pthread_self());
    expires = 0;
    std::string stmt = "UPDATE register SET time=current_timestamp WHERE client_id=$1";
    std::vector<std::string> params { std::to_string(id) };
    CaDbRes r = d->dbh->execute(stmt, params);
    time(&expires);
    expires += d->cli_ttl;
    if(!r.error() && r.affected_rows == 0) {
        d->err = true;
        d->msg = "ca-registrar: no client with id " + std::to_string(id) + " recorded into database";
    }
    else {
        d->err = r.error();
        d->msg = r.errmsg;
    }
    return d->err ? 0 : id;
}

bool CaRegistrar::m_record_up_to_date(const unsigned long &id, CaDbH *dbh)  const {
    assert((d->this_thread == pthread_self() && dbh == d->dbh) || dbh == d->dbh2);
    auto start = std::chrono::high_resolution_clock::now();
    std::string stmt = "SELECT time FROM register WHERE client_id=$1 AND time > "
            " current_timestamp - interval ' " + std::to_string(d->cli_ttl) + " second' ";
    std::vector<std::string> params { std::to_string(id) };
    CaDbRes r = dbh->execute(stmt, params);
    d->err = r.error() || r.size() > 1;
    auto end = std::chrono::high_resolution_clock::now();
    printf("CaRegistrar::m_record_up_to_date took %ldus and id %ld %s up to date.\n",
           std::chrono::duration_cast<std::chrono::microseconds>(end - start).count(), id, (!d->err && r.size() == 1) ? "\e[1;32mIS\e[0m" : "\e[1;31mIS NOT\e[0m");
    if(!d->err && r.size() == 1)
        return true;
    else if(r.size() > 1)
        d->msg = "ca-registrar.m_record_up_to_date: unexpected number of records for id " + std::to_string(id);
    else
        d->msg = r.errmsg;
    return !d->err && r.size() == 1;
}

bool CaRegistrar::m_delete_expired_records() {
    assert(d->this_thread == pthread_self());
    std::string stmt = "DELETE FROM register WHERE time < current_timestamp - interval ' "
            + std::to_string(d->cli_ttl) + " second' ";
    CaDbRes r = d->dbh->execute(stmt, std::vector<std::string>());
    // clients table is ON DELETE CASCADE on client id
    d->err = r.error();
    d->msg = r.errmsg;
    return !d->err;
}

CuData CaRegistrar::m_make_ok_ttl_id(time_t &ttl, const unsigned long long &id) const
{
    CuData r("exp", ttl);
    r["id"] = id;
    r["ttl"] = d->cli_ttl;
    return r;
}

bool CaRegistrar::m_get_up_to_date_records(CuData& out) {
    assert(d->this_thread == pthread_self());
    std::vector<std::string> v;
    d->err = !d->dbh->connected;
    if(!d->err) {
        d->msg.clear();
        std::string stmt = "SELECT client_id FROM register WHERE time > current_timestamp - interval ' "
                + std::to_string(d->cli_ttl) + " second' ";
        CaDbRes r = d->dbh->execute(stmt, std::vector<std::string>());
        // assume id from db is a valid integer, do not try/catch
        for(size_t i = 0; !r.error() && i < r.size(); i++)
            v.push_back(r.value(i, "client_id"));
        d->err = r.error();
        if(d->err) d->msg = r.errmsg;
    }
    else
        d->msg = "ca-registrar.m_get_up_to_date_records: database not connected: " + d->dbh->msg;
    out["uptodate"] = v;
    return !d->err;
}

bool CaRegistrar::m_delete_unsubscribed(const unsigned long long &id) {
    assert(d->this_thread == pthread_self());
    d->err = !d->dbh->connected;
    if(!d->err) {
        d->msg.clear();
        std::string stmt = "DELETE FROM register WHERE client_id=$1";
        CaDbRes r = d->dbh->execute(stmt, std::vector<std::string>{std::to_string(id)});
        d->err = r.error();
        if(d->err) d->msg = r.errmsg;
    }
    return !d->err;
}

bool CaRegistrar::m_get_srv_id() {
    assert(d->this_thread == pthread_self());
    CuData connp(d->connp); // thread local copy
    std::string conf_id;
    d->srv_id = -1;
    d->err = !d->dbh->connected;
    if(!d->err) {
        d->msg.clear();
        std::string stmt = "SELECT conf_id FROM srvconf WHERE srvconf.addr=$1 AND srvconf.port=$2";
        const std::string& h = connp.s("async-server-address"), &p = connp.s("async-server-port");
        std::vector<std::string> params { h, p };
        CaDbRes r = d->dbh->execute(stmt, params);
        if(!r.error() && r.size() == 0) { // no configuration: insert the new one
            stmt = "INSERT INTO srvconf (addr,port) VALUES ($1,$2) RETURNING conf_id";
            r = d->dbh->execute(stmt, std::vector<std::string>{ h, p });
            if(!r.error() && r.size() == 1)
                conf_id = r.value(0, "conf_id");
        }
        else if(!r.error() && r.size() == 1) {
            conf_id = r.value(0, "conf_id");
        }
        if(!r.error() && conf_id.length() > 0) { // do not try catch, conf_id is integer
            stmt = "SELECT id FROM service WHERE conf_id=$1";
            r = d->dbh->execute(stmt, std::vector<std::string> { r.value(0, "conf_id") });
            if(!r.error() && r.size() == 1)
                d->srv_id = atol(r.value(0, "id").c_str());
            else if(r.size() == 0) {
                stmt = "INSERT INTO service (id,started,expected,conf_id) "
                    " VALUES(DEFAULT,NULL,NULL,$1) RETURNING id";
                r = d->dbh->execute(stmt, std::vector<std::string>{ conf_id } );
                if(!r.error() && r.size() > 0)
                    d->srv_id = atol(r.value(0, "id").c_str());
            }
        }
        d->err = r.error();
        if(d->err) {
             d->msg = "unexpected database error while executing '" + stmt + ": \"" + r.errmsg + "\"";
        }
    }
    return !d->err;
}

int CaRegistrar::getType() const {
    return CaRegistrarType;
}

void CaRegistrar::event(CuActivityEvent *) { }

bool CaRegistrar::matches(const CuData &token) const {
    return token == getToken();
}

int CaRegistrar::repeat() const {
    return -1;
}

void CaRegistrar::init() {
    CuData conp(d->connp); // thread local copy
    d->dbh = new CaDbH(conp.s("dbuser"), conp.s("dbpass"), conp.s("dbnam"), conp.s("dbhost"), conp.s("dbport"));
    if(!d->dbh->connect()) publishResult(CuData("err", true).set("msg", d->dbh->msg));
    else {
        d->this_thread = pthread_self();
        bool ok = m_get_srv_id();
        publishResult(ok  ? CuData("msg", d->dbh->msg).set("srv_id", d->srv_id) : CuData("err", true).set("msg", d->msg));
    }
}

// executes the operation indicated in the events in the queue
void CaRegistrar::execute() {
    if(d->dbh->connected) {
        while(!d->quit) {
            std::list<const CaRegiEvent *> events;
            {
                std::unique_lock<std::mutex> lock(d->mu);
                // copy messages locally while holding lock
                while(d->eq.empty() && !d->quit)
                    d->wait.wait(lock);
                while(!d->eq.empty() && !d->quit) {
                    events.emplace_back(std::move(d->eq.front()));
                    d->eq.pop();
                }
            }// end locked section
            // process messages
            unsigned long long id = 0;
            time_t expires = 0;
            for(const CaRegiEvent * xd : events) {
                if(xd->type == CaRegiEvent::RECheckExpired) {
                    CuData x;
                    if(!m_get_up_to_date_records(x)) // x will have "expired" key
                        x.set("err", true).set("msg", d->msg);
                     else
                       d->err = !m_delete_expired_records();
                    if(!x.isEmpty())
                        publishResult(x);
                }
                else if(xd->type == CaRegiEvent::REUnsubscribe) {
                    d->err = m_delete_unsubscribed(xd->tok);
                    publishResult(CuData("u", std::vector<std::string>{ std::to_string(xd->tok)}).set("fd", xd->fd));
                }
                else if(xd->type == CaRegiEvent::RERecord) {
                    if(xd->tok == 0)
                        id = m_insert_record(xd->host, expires);
                    else if(xd->src.length() == 0)
                        id = m_update_record(xd->tok, xd->host, expires);

                    if(d->err) {
                        publishResult(CuData("err", true).set("msg", d->msg).set("fd", xd->fd));
                    }
                    else if(expires > 0 && id > 0) {
                        publishResult(m_make_ok_ttl_id(expires, id).set("fd", xd->fd));
                    }
                }
                delete xd;
            }
        }
    }
}

void CaRegistrar::onExit() {
    if(d->dbh) delete d->dbh;
}
