#include "caproxy.h"
#include "caproxyhttph.h"
#include "caproxyjson_h.h"
#include "caunsubmulticast.h"
#include "caprosock.h"
#include "ca-receiver-a.h"
#include "cahttpmsg.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h> // inet_ntop
#include <time.h>
#include <atomic>

#include <cudata.h>
#include <cahttpmsg_u.h>
#include <cajson-data-encoder.h>
#include <vector>
#include <mutex>
#include <shared_mutex>
#include <condition_variable>
#include <queue>

#include <assert.h>
#include <future>

#define _BUFSIZ 512

/*!
 * \brief build an event for proxying a message
 * \param _clifd the client socket file descriptor
 * \param httpt the http type code
 * \param _http_p the http payload
 * \param _type either ActionProxySync or ActionProxyAsync to identify the caserver
 *        to forward msg (either sync or async respectively)
 */
CaProxyEvent::CaProxyEvent(int _clifd, CaPHttpMsgType httpt, const std::string &_paylo, const CaProxyRoute& pr, CaProxyEventType t)
    : cli_fd(_clifd), http_p(_paylo), route(pr), ety(t), httpmty(httpt) {
}

/*!
 * \brief send a message to a caserver: sync or async according to CaProxy::CaProActionType
 * \param httpt the http type code
 * \param _http_p http payload
 * \param _type either ActionProxySync or ActionProxyAsync
 */
CaProxyEvent::CaProxyEvent(CaPHttpMsgType httpt, const std::string &_payl, const CaProxyRoute &pr, CaProxyEventType t)
    : cli_fd(-1), http_p(_payl), route(pr), ety(t), httpmty(httpt) {
}

CaProxyEvent::CaProxyEvent(int _clifd,
                           CaPHttpMsgType _httpty,
                           const std::string &syr,
                           const std::string &s_asy,
                           const std::string &u_asy,
                           const std::string &httphdr,
                           unsigned long orig_clen)
    : cli_fd(_clifd), sy_r(syr), s_asy_r(s_asy), u_asy_r(u_asy), httpmty(_httpty), ety(CaPEMixed), http_hdr(httphdr), clen0(orig_clen) {
}

class CaProxyPrivate {
public:
    CuData connp;
    long srv_id;
    CaProxySo sy_sock, asy_sock, sup_sock;
    size_t id; // id of thread in the pool
    std::string srvaddr, supervi_errmsg; // supervisor error messages are not fatal
    unsigned short srvport;
    unsigned upstream_timeo;
    CaHttpMsgU http_u;


    std::queue <CaProxyEvent *> eq;
    std::mutex mu;
    std::condition_variable wait;
    bool quit, running, supv_en;
    std::atomic<bool> busy;

    CaUnsubMulticast* unsub_multi;
    std::vector<std::string> errors;
    CaReceiver_A* receiver_a;

    pthread_t xthread;
};

CaProxy::CaProxy(const CuData& opts,
                 const std::string& servaddr,
                 const unsigned short servport,
                 int id,
                 long srv_id,
                 const CuData& token,
                 CaReceiver_A *rec_a)
    : CuActivity(token)
{
    d = new CaProxyPrivate;
    d->id = id;
    d->srv_id = srv_id;
    const std::vector<std::string>& pa { "sync-server-address", "sync-server-port", "async-server-address", "async-server-port",
                                         "supervisor-server-address", "supervisor-server-port" };
    for(const std::string& p : pa)
        d->connp[p] = opts[p];
    d->srvaddr = servaddr;
    d->srvport = servport;
    d->upstream_timeo = opts.containsKey("upstream-timeo") ? opts.U("upstream-timeo") : 0U;

    d->quit = d->running = d->busy = false;
    d->supv_en = opts.containsKey("supervisor-enable") && opts.B("supervisor-enable") && d->connp.containsKey("supervisor-server-address") && d->connp.containsKey("supervisor-server-port");
    d->unsub_multi = new CaUnsubMulticast(opts);
    d->receiver_a = rec_a;
    setFlag(CuActivity::CuADeleteOnExit, true);
    d->xthread = 0;
}

CaProxy::~CaProxy() {
    printf("CaProxy::~CaProxy %p\n", this);
    delete d->unsub_multi; // waits for threads
    delete d;
}

void CaProxy::proxy_event(CaProxyEvent *e) {
    std::unique_lock<std::mutex> lock(d->mu);
    d->eq.push(e);
    d->wait.notify_one();
}

void CaProxy::execute() {
    d->xthread = pthread_self();
    while(!d->quit) { // wait for events
        d->running = true;
        d->busy = false;
        std::list<const CaProxyEvent *> events;
        {
            std::unique_lock<std::mutex> lock(d->mu);
            // copy messages locally while holding lock
            while(d->eq.empty() && !d->quit)
                d->wait.wait(lock);
            while(!d->eq.empty() && !d->quit) {
                events.emplace_back(std::move(d->eq.front()));
                d->eq.pop();
            }
        }// end locked section
        // process messages
        d->busy = true;

        CaProxyReply preply;
        std::string reply, errormsg;
        for(const CaProxyEvent * xd : events) {
            d->errors.clear(); // clear errors
            if(xd->cli_fd > -1 && xd->ety == CaProxyEvent::CaPEMixed) {
                preply = do_proxy(xd->cli_fd, xd->httpmty, xd->sy_r, xd->s_asy_r, xd->u_asy_r, xd->http_hdr, xd->clen0);
            }
            else if(xd->cli_fd < 0) {
                printf("[0x%lx] CaProxy::execute: cli fd < 0 calling write_read_caserver with %d %s %d HEADER %s\e[0m\n",
                       pthread_self(),  xd->route.proxy_to, xd->http_p.c_str(), xd->ety, xd->http_hdr.c_str());
                preply = write_read_caserver(xd->route, xd->httpmty, xd->http_p, xd->http_hdr, xd->clen0, xd->ety);
                printf("[0x%lx] CaProxy.execute: write_read_caserver reply \"%s\" fwd %d rep mode %d\n", pthread_self(),  reply.c_str(),
                       xd->route.proxy_to, xd->route.reply_mode);
            }
            else if(xd->route.proxy_to == CaProxyRoute::ToClient) // send message to client
                preply = msg_send_client(xd->cli_fd, xd->httpmty, xd->http_p);
            if(preply.error.length() > 0)
                publishResult(CuData("err", true).set("msg", preply.error));
            if(xd->cli_fd > 0)
                d->receiver_a->sock_close(xd->cli_fd);
            delete xd;
        }
    }
    d->running = false;
}

// fd: file descriptor of the client waiting for a reply
// syr json sync request
// asyr json async subscribe request
// uasy json async unsubscribe request
// d->errors is not cleared by this function
//
// For each request (syr, sasy, uasy) an async task is started and executes
// write_read_caserver
// If a network error occurs while connecting to either caserver sync or async,
// the client will receive a 502 response and the http payload will contain the
// error message(s) only.
// For example, if sync succeeds while async fails, reply from sync is discarded,
// the HTTP response status is set to 502 (bad gateway, i.e. bad caserver sync) and
// the HTTP payload will store the connection error message, e.g.:
//
//    [{"err":true,"msg":"caserver async socket connection error: 192.168.1.159:9294: Connection refused"}]
//
// If both caserver fail, 502 message will bear the two error messages from the servers in the payload
// (in a JSON array):
//
//    [{"err":true,"msg":"caserver sync socket connection error: 192.168.1.159:9293: Connection refused"},{"err":true,"msg":"caserver async socket connection error: 192.168.1.159:9294: Connection refused"}]
//
// If there is no error, all replies obtained by caservers are merged together in an array
// For example, a subscribe request for n sources will bear the first n results from the sync caserver
// plus a final element describing the caserver async result, of the type:
//
//    {"count":2,"id":270,"msg":"ok","type":"async"}
//
CaProxyReply CaProxy::do_proxy(int fd,
                               const CaProxyEvent::CaPHttpMsgType httpty, // from CaProxyProcess::m_data_redistrib header = line1 + http hdr
                               const std::string &syr, // json sync req
                               const std::string& sasy, // json async sub req
                               const std::string& uasy,
                               const std::string& http_hdr,
                               unsigned long clen0) {
    assert(d->xthread == pthread_self());
    CaProxyReply ret_reply;
    CaProxyJsonH jsh;
    std::string supj, errmsg;
    CaProxyJsonH jh;
    CaProxyReply asyrep, syrep, asyurep;
    int httpstatus = -1;

    if(syr.length() > 0) { // send to sync caserver
        syrep = write_read_caserver(CaProxyRoute(CaProxyRoute::ToSync), httpty, syr, http_hdr, clen0, CaProxyEvent::CaPESub);
    }
    if(sasy.length() > 0) {
        asyrep = write_read_caserver(CaProxyRoute(CaProxyRoute::ToAsync), httpty, sasy, http_hdr, clen0, CaProxyEvent::CaPESub);
    }
    if(uasy.length() > 0) {
        asyurep = write_read_caserver(CaProxyRoute(CaProxyRoute::ToAsync), httpty, uasy, http_hdr, clen0, CaProxyEvent::CaPEUnsub);
    }

    std::string_view syjs, asyjs, asyujs;
    // network error ?
    // sync
    bool errors = syrep.error.length() > 0 || asyrep.error.length() > 0 || asyurep.error.length() > 0;
    std::string errs[3];
    if(errors) {
        if(syrep.error.length() > 0)
            errs[0] = jh.make_err(syrep.error);
        if(asyrep.error.length() > 0)
            errs[1] = jh.make_err(asyrep.error);
        if(asyurep.error.length() > 0)
            errs[2] = jh.make_err(asyurep.error);
        syjs = errs[0]; asyjs = errs[1]; asyujs = errs[2];
    }

    CaHttpMsg ms, mw, mu;
    if(errors) {
        httpstatus = 502;
    }
    else {
        // syrep.reply, asyrep.reply and asyurep.reply are full HTTP messages
        // use CaHttpMsg to get the respective JSON payloads: syjs, asyjs, asyujs
        //
        if(!syrep.is_empty()) {
            ms.from(syrep.reply);
            syjs = ms.payload(); // cp from string_view. nlohmann_json doesn't work with strings
            if(httpstatus < syrep.code)
                httpstatus = syrep.code;
        }
        if (!asyrep.is_empty()) {
            mw.from(asyrep.reply);
            asyjs = mw.payload();
            if(httpstatus < syrep.code)
                httpstatus = syrep.code;
        }
        // async unsubscribe
        if (!asyurep.is_empty() && !errors) {
            mu.from(asyurep.reply);
            asyujs = mu.payload();
            if(httpstatus < syrep.code)
                httpstatus = syrep.code;
        }
    }

    // merge
    ret_reply.reply = jh.merge(syjs, asyjs, asyujs, errmsg); // merge sets errmsg to some error message if json fails

    if(d->errors.size() == 0 && errmsg.length() > 0) {
        ret_reply.reply = jh.make_err(errmsg); // no network error, but error from jh
        d->errors.push_back(errmsg);
    }

    //    CaProxySoResult res = d->errors.size() == 0 ? CaPSoOk : CaPCaservSoErr;
    // error from caserver: prepare a Json error response for the client

    // ----------------------------------
    // proxying: write back to *client*
    // ----------------------------------

    // 502 bad gateway in case of net errors to upstream, 400 in case of error

    CaHttpMsg m;
    m.make_reply(httpstatus > 0 ? httpstatus : 200, ret_reply.reply, httpstatus > 0);
    int bw = CaProSock().sock_write(fd, m.buf.c_str(), errmsg);
    printf("[0x%lx] [%p]\e[0;33m <<< out <<< \e[0mcaproxy.do_proxy wrote %d bytes to fd %d error %s this %p d HTTP STATUS %d\e[0m\n",
           pthread_self(),this, bw, fd, errmsg.c_str(),  d, httpstatus);
    // res is not used
    //    res = bw > 0 ? CaPSoOk : CaPCliSoErr;
    // client closes the connection: do not close file descriptor fd
    if(errmsg.length() > 0)
        d->errors.push_back(errmsg);

    // -----------------------------------
    // supervisor
    // -----------------------------------
    //
    // after either a successful async subscribe or unsubscribe (or both), forward the request to
    // the supervisor. supfwd returns true if CaPEUnsub or (CaPESub && reply without errors)
    // in case of CaPEUnsub forward the unsubscribe to the supervisor regardless errors from the caserver
    if(asyjs.length() > 0 && d->supv_en && supfwd(asyjs, CaProxyEvent::CaPESub) /*&& jsh.insert_srvid_evtyp(sasy, d->srv_id, supj, CaProxyEvent::CaPESub)*/) {
        write_read_supervisor(m.build(http_hdr, sasy).buf, errmsg); // insert srv_id in json
    }
    if(asyujs.length() > 0 && d->supv_en && supfwd(asyujs, CaProxyEvent::CaPEUnsub) /*&& jsh.insert_srvid_evtyp(uasy, -1, supj, CaProxyEvent::CaPEUnsub)*/) {
        write_read_supervisor(m.build(http_hdr, uasy).buf, errmsg); // srv_id -1
    }

    return ret_reply;
}

/*!
 * \brief send msg to either sync or async caserver
 * \param at type of caserver (picks async or sync socket)
 * \param httpm complete HTTP message buffer
 * \return CaProxyReply CaProxyReply.reply will contain the reply from the server if no error
 *         occurred. In case of errors, CaProxyReply.error will store the error message
 *
 * if previous socket fd is non negative and there's a connection error, reconnection is attempted
 *
 * Unsubscribe requests are broadcasted to _all_ caservers through
 * CaUnsubMulticast::send_multi (see README.md)
 *
 */
CaProxyReply CaProxy::write_read_caserver(const CaProxyRoute & proute,
                                          const CaProxyEvent::CaPHttpMsgType httpt,
                                          const std::string &paylo,
                                          const std::string& hdr,
                                          const int orig_clen,
                                          CaProxyEvent::CaProxyEventType ety) {
    assert(d->xthread == pthread_self());

    CaProSock psock;
    CaProxyReply preply(200); // init with OK code
    int nread = -1, nwrit = -1;
    int retry = 0;
    CaProxySo ssock = proute.proxy_to == CaProxyRoute::ToSync ? d->sy_sock : (ety == CaProxyEvent::CaPESub ? d->asy_sock : CaProxySo {0, ""});
    // build a CaHttpMsg with hdr and payload. orig_clen is used to skip header mangling when equal to payload length
    const CaHttpMsg m = hdr.length() > 0 ? std::move(CaHttpMsg(hdr, paylo, orig_clen)) : std::move(CaHttpMsg(CaHttpMsg::Post, "src-bundle", paylo));
    errno = 0;
    //    printf("[0x%lx] CaProxy::write_read_caserver enter: proute.fwd %d event ty %d \e[1;3;33mssock.sofd %d\e[0m ROUTE %s\e[0m\n",
    //           pthread_self(), proute.fwd, ety, ssock.sofd, proute.fwd == CaProxyRoute::ToSync ? "\e[1;32mTO SYNC" : "\e[1;31mOTHER");

    if(ssock.sofd < 0) { // try reconnect. If ssock is 0 then we need fwd multiple async (unsubscribe)
        (proute.proxy_to == CaProxyRoute::ToSync)  ? ssock = m_sy_connect() : ssock = m_asy_connect();
        if(ssock.sofd < 0)
            preply.setError(ssock.errmsg).setCode(502);
    }
    while(ety != CaProxyEvent::CaPEUnsub && ssock.sofd > 0 && retry++ <= 2 && nread <= 0) {
        // build a CaHttpMsg with hdr and payload. orig_clen is used to skip header mangling when equal to payload length
        const CaHttpMsg m(hdr, paylo, orig_clen);
        std::string addr;
        // even if ssock is > 0 because an earlier connection attempt succeeded, the corresponding socket may have
        // gone down and up again. If write fails on ssock, try to reopen it
        nwrit = CaProSock().sock_write(ssock.sofd, m.buf.data(), preply.error);
        // caserver will reply something that must be sent back on fd
        if(ssock.sofd > -1 && nwrit > 0) {
            nread = psock.sock_read(ssock.sofd, preply.reply, preply.error);
        }
        // otherwise nread remains negative
        if(nread <= 0) { // try reconnect
            printf("[0x%lx] CaProxy::write_read_caserver read failed (nread=%d) on fd %d: reconnecting\n", pthread_self(), nread, ssock.sofd);
            m_sock_close(ssock.sofd, proute.proxy_to); // close ssock and reset d->asy_sock or d->sy_sock to -1
            (proute.proxy_to == CaProxyRoute::ToSync)  ? ssock = m_sy_connect() : ssock = m_asy_connect();
            if(ssock.sofd < 0) // reconnection error!
                preply.setError(ssock.errmsg).setCode(502);
        }
        printf("[0x%lx] CaProxy::write_read_caserver this %p WRITE OK %d bytes: %s READ OK %d BYTES %s \e[1;3;33mssock.sofd %d\e[0m\n",
               pthread_self(), this, nwrit, nwrit > 0 ? "\e[1;32mYES\e[0m" : "\e[1;31mNO\e[0m",
               nread, nread > 0 ? "\e[1;32mYES\e[0m" : "\e[1;31mNO\e[0m", ssock.sofd);
    }
    if(ety == CaProxyEvent::CaPEUp2Date && proute.proxy_to == CaProxyRoute::ToAsync && d->supv_en) {
        // client id expired (client vanished): send to supervisor and all async caservers
        // no reply to the (expired) client
        // coming here from caregistrarlistener via proxy pool -> msg_send_caserver
        // send the unsubscribe event to the supervisor even
        // if the ensuing notification to che caservers fails
        std::string superr;
        CaHttpMsg m;
        m.make_post("src-bundle", paylo, true);
        write_read_supervisor(m.buf, superr); // srv_id -1: unknown because u is broadcasted
    }

    if((ety == CaProxyEvent::CaPEUnsub || ety == CaProxyEvent::CaPEUp2Date) && proute.proxy_to == CaProxyRoute::ToAsync) {
        CaHttpMsg m;
        m.make_post("src-bundle", paylo, true);
        // unsubscribe must be sent to all known async servers, listed in the configuration file
        printf("caproxy.cpp: sending multicast unsub \e[0;36m%s\e[0m -- hdr was \e[1;33m%s\e[0m\n", m.buf.c_str(), hdr.c_str());
        preply.code = d->unsub_multi->send_multi(m.buf, preply.reply);
    }
    //    printf("[0x%lx] CaProxy::write_read_caserver proute.fwd %d event ty %d reply len \e[1;32m%ld\e[0m err \e[1;31m%s\e[0m SOCK FD %d\n",
    //           pthread_self(), proute.fwd, ety, preply.reply.length(), preply.error.c_str(), ssock.sofd);
    return preply;
}

// write msg to supervisor
// save errors into d->supervi_errmsg
// d->errmsg left untouched so that errors here are not fatal and not reported to the client
bool CaProxy::write_read_supervisor(const std::string &msg, std::string& errmsg) {
    assert(d->xthread == pthread_self());
    int nread = -1, nwrit = -1;
    int retry = 0;
    errno = 0;
    errmsg = std::string();
    std::string casup_reply;
    CaProSock pso;
    if(d->sup_sock.sofd < 0) // try reconnect
        d->sup_sock = m_supervi_connect();

    // same logic as write_read_caserver
    while(d->sup_sock.sofd > 0 && retry++ <= 2 && nread <= 0) {
        nwrit = pso.sock_write(d->sup_sock.sofd, msg.c_str(), errmsg);
        // caserver will reply something that must be sent back on fd
        if(d->sup_sock.sofd > -1 && nwrit > 0) {
            nread = pso.sock_read(d->sup_sock.sofd, casup_reply, errmsg);
        }
        // otherwise nread remains negative
        if(nread <= 0) { // try reconnect
            m_sock_close(d->sup_sock.sofd, CaProxyRoute::ToSupervisor); // close ssock and reset d->asy_sock or d->sy_sock to -1
            d->sup_sock = m_supervi_connect();
        }
    }
    return nwrit > 0 && nread > 0;
}

void CaProxy::stop() {
    d->quit = true;
    d->wait.notify_one();
}

std::string CaProxy::errmsg() const {
    std::string e;
    for(size_t i = 0; d->errors.size() > 0 && i < d->errors.size() - 1; i++)
        e += d->errors[i] + ", ";
    if(d->errors.size() > 0)
        e += d->errors[d->errors.size() - 1];
    return e;
}

bool CaProxy::sy_connected() const {
    return d->sy_sock.sofd > -1;
}

bool CaProxy::asy_connected() const {
    return d->asy_sock.sofd > -1;
}

bool CaProxy::supervisor_connected() const {
    return d->sup_sock.sofd > -1;
}


CaProxySo CaProxy::m_sy_connect() {
    assert(d->xthread == pthread_self());
    CuData conp(d->connp); // thread local copy
    d->sy_sock.sofd = CaProSock().sock_connect(conp.s("sync-server-address").c_str(), conp.s("sync-server-port").c_str(), d->sy_sock.errmsg, d->upstream_timeo);
    if(d->sy_sock.sofd < 0)
        d->sy_sock.errmsg = "caserver sync socket connection error: " + d->sy_sock.errmsg; // improve message
    return d->sy_sock;
}

CaProxySo CaProxy::m_asy_connect() {
    assert(d->xthread == pthread_self());
    CuData conp(d->connp); // thread local copy
    d->asy_sock.sofd = CaProSock().sock_connect(conp.s("async-server-address").c_str(), conp.s("async-server-port").c_str(), d->asy_sock.errmsg, d->upstream_timeo);
    if(d->asy_sock.sofd < 0)
        d->asy_sock.errmsg = "caserver async socket connection error: " + d->asy_sock.errmsg;
    return  d->asy_sock;
}

// connect to supervisor
// errors saved in d->supervi_errmsg
// d->errmsg left untouched: supervisor connection errors are not sent back to client
CaProxySo CaProxy::m_supervi_connect()
{
    assert(d->xthread == pthread_self());
    CuData conp(d->connp); // thread local copy
    d->sup_sock.sofd = CaProSock().sock_connect(conp.s("supervisor-server-address").c_str(), conp.s("supervisor-server-port").c_str(), d->sup_sock.errmsg, d->upstream_timeo);
    if(d->sup_sock.sofd < 0)
        d->sup_sock.errmsg = "ca-supervisor connection error: " + d->sup_sock.errmsg;
    return  d->sup_sock;
}

CaProxyReply CaProxy::msg_send_client(int fd, const CaProxyEvent::CaPHttpMsgType httpt, const std::string &http_p) {
    assert(d->xthread == pthread_self());
    CaHttpMsg m(httpt, http_p, true); // empty X-Source, connection: close
    CaProxyReply r;
    CaProSock().sock_write(fd, m.buf.data(), r.error);
    return r;
}


bool CaProxy::is_running() const {
    return d->running;
}

bool CaProxy::is_busy() const {
    std::unique_lock<std::mutex> lock(d->mu);
    return d->busy || d->eq.size() > 0;
}

size_t CaProxy::id() const {
    return d->id;
}

void CaProxy::set_id(size_t id) {
    d->id = id;
}

void CaProxy::m_sock_close(int sofd, const CaProxyRoute& prou)
{
    assert(d->xthread == pthread_self());
    if(prou.proxy_to == CaProxyRoute::ToSync) d->sy_sock = CaProxySo();
    else if(prou.proxy_to == CaProxyRoute::ToAsync) d->asy_sock = CaProxySo();
    else if(prou.proxy_to == CaProxyRoute::ToSupervisor) d->sup_sock = CaProxySo();
    if(close(sofd) < 0) {
        d->errors.push_back("caproxy.m_sock_close: error closing socket fd " + std::to_string(sofd) + ": " + strerror(errno));
        perr("%s", d->errors.at(d->errors.size() - 1).c_str());
    }
    else
        printf("\e[0;32mCaProxy::m_sock_close: closed socket %d\e[0m\n", sofd);
}

/*!
 * \brief determine whether the message msg shall be forwarded to the supervisor
 * \param asyrep the async reply received from caserver async
 * \param ety the event type, used to distinguish between subscribe and unsubscribe operations
 * \return true if 1. the event type is unsubscribe or 2. the request is subscribe and asyrep contains "msg":"ok"
 *
 * unsubscribes are always forwarded to ca-supervisor in order to have activities removed from the database
 * subscribes are forwarded to ca-supervisro only if caserver async replies "ok"
 */
bool CaProxy::supfwd(const std::string_view &json, CaProxyEvent::CaProxyEventType ety) const {
    if(ety == CaProxyEvent::CaPEUnsub)
        return true;
    CaProxyJsonH jsh;
    printf("CaProxy::supfwd reply ok? %s type %d\n", json.data(), ety);
    bool is_ok = jsh.json_reply_ok(json);
    return ety == CaProxyEvent::CaPESub && is_ok;
}

void CaProxy::set_srv_id(long id) {
    d->srv_id = id;
}

int CaProxy::getType() const {
    return CaProxyAType;
}

bool CaProxy::matches(const CuData &token) const {
    return getToken() == token;
}

int CaProxy::repeat() const {
    return -1;
}

void CaProxy::event(CuActivityEvent *e) {

}

void CaProxy::init() {
}

void CaProxy::onExit() {
}
