#ifndef CAPROXY_H
#define CAPROXY_H

#include <string>
#include <cuactivity.h>
#include <cuactivityevent.h>

class CaProxyPrivate;
class CuData;
class CaProxyEvent;
class CaReceiver_A;

class CaProxyRoute {
public:
    /*! forward to sync caserver, async caserver, or generic message to a
     *  specified file descriptor
     */
    enum CaP_ProxyTo { ToUndefined = -1, ToSync = 1, ToAsync = 0x2, ToClient = 0x4, ToSupervisor = 0x8 };

    /*! Reply: proxy the reply back to the client
        Sink: sink the reply here
     */
    enum CaP_ReplyMode { Reply, None };

    CaProxyRoute(CaP_ProxyTo fw, CaP_ReplyMode rep_m = Reply)
        : proxy_to(fw), reply_mode(rep_m) {}

    CaProxyRoute() : proxy_to(ToUndefined), reply_mode(None) {}


    CaP_ProxyTo proxy_to;
    CaP_ReplyMode reply_mode;
    bool to_caserver() const { return proxy_to == ToSync || proxy_to == ToAsync; }
};

class CaProxyEvent  {
public:
    enum CaProxyEventType { CaPEUndef = 0, CaPESub = 1, CaPEUnsub = 0x2, CaPESync = 0x4, CaPEMixed = 0x8, CaPEUp2Date = 0x10 };
    enum CaPHttpMsgType { CaPHttpPost = 0, CaPHttpReplyOk = 200,
                          CaPHttp400 = 400, CaPHttp402 = 402,
                          CaPHttp500 = 500, CaPHttp502 = 502, CaPHttp503 = 503 };

    CaProxyEvent(int _clifd, CaPHttpMsgType httpt, const std::string &http_p, const CaProxyRoute &pr, CaProxyEventType t);
    CaProxyEvent(CaPHttpMsgType httpt, const std::string &_payl, const CaProxyRoute &pr, CaProxyEventType t);
    CaProxyEvent(int _clifd,
                 CaPHttpMsgType httpt,
                 const std::string& syr,
                 const std::string& s_asy,
                 const std::string& u_asy,
                 const std::string& httphdr, // original http header and clen0 its length
                 unsigned long clen0);

    int cli_fd;
    std::string http_p;
    std::string sy_r, s_asy_r, u_asy_r;
    std::string http_hdr;
    CaProxyRoute route;
    CaProxyEventType ety;
    CaPHttpMsgType httpmty;
    unsigned long clen0;
};

class CaProxySo {

public:
    int sofd = -1;
    std::string errmsg;
};

class CaProxyReply {
public:
    CaProxyReply()  : code(-1) {}
    CaProxyReply(int co) : code(co) {}

    CaProxyReply& setError(const std::string& err) {
        error = err;
        return *this;
    }

    CaProxyReply& setCode(int c) {
        code = c;
        return *this;
    }

    CaProxyReply& setReply(const std::string& buf) {
        reply = buf;
        return *this;
    }

    std::string reply; // json reply
    std::string error; // error msg
    int code;          // http suggested code

    bool is_empty() const { return reply.length() == 0 && error.length() == 0; }
};


class CaProxy : public CuActivity
{
public:
    enum CaProxyActivityType { CaProxyAType = 2022 };
    enum CaProxySoResult { CaPSoOk, CaPCaservSoErr, CaPCliSoErr };

    CaProxy(const CuData& opts,
            const std::string& servaddr,
            const unsigned short servport,
            int id,
            long srv_id,
            const CuData &token,
            CaReceiver_A *rec_a);
    virtual ~CaProxy();
    std::string errmsg() const;
    bool sy_connected() const;
    bool asy_connected() const;
    bool supervisor_connected() const;
    CaProxyReply do_proxy(int fd,
                          const CaProxyEvent::CaPHttpMsgType httpty,
                          const std::string &syr,
                          const std::string &sasy,
                          const std::string &uasy, const std::string &x_source,
                          unsigned long clen0);

    CaProxyReply msg_send_client(int fd,
                          const CaProxyEvent::CaPHttpMsgType httpt,
                          const std::string& http_p);

    CaProxyReply write_read_caserver(const CaProxyRoute &at,
                                     const CaProxyEvent::CaPHttpMsgType httpt,
                                     const std::string& http_payload,
                                     const std::string &http_hdr,
                                     const int orig_clen,
                                     CaProxyEvent::CaProxyEventType ety);

    bool write_read_supervisor(const std::string& msg, std::string &errmsg);
    bool is_running() const;
    bool is_busy() const;
    void set_id(size_t id);
    size_t id() const;
    void stop();
    bool supfwd(const std::string_view &json, CaProxyEvent::CaProxyEventType ety) const;
    void set_srv_id(long id);

    // add event to queue
    void proxy_event(CaProxyEvent *e);

private:
    CaProxySo m_sy_connect();
    CaProxySo m_asy_connect();
    CaProxySo m_supervi_connect();

    void m_sock_close(int sofd, const CaProxyRoute &prou);

    CaProxyPrivate *d;
    std::string m_make_read_err(const std::string& payload);

    // CuActivity interface
public:
    int getType() const;
    bool matches(const CuData &token) const;
    int repeat() const;
    void event(CuActivityEvent *e);

protected:
    void init();
    void execute();
    void onExit();
};



#endif // CAPROXY_H
