#include "caproxyjson_h.h"
#include <nlohmann/json.hpp>
#include <cumacros.h>
#include <cudata.h>
#include <stdexcept>
#include <set>
#include <caidlisth.h>

CaProxyJsonH::CaProxyJsonH()
{

}

/*! divide the request bundle separating entries into
 1. "s" subscribe requests shall be sent to both sync and async caservers.
 The sync module replies to client, while the async reply is received by the
 proxy.
 2. "r", "dbr", "conf", ... sync requests that shall be forwarded to the sync caserver only
 Replies are proxied back to the client
 3. "u": unsubscribe: it is directed to caproxy "async" only and in this case the caproxy
 async reply shall be sent to the client.

 since the bundle may in principle contain different kinds of requests, we must divide them
 in three groups as above
*/
bool CaProxyJsonH::divide(const std::string &rin,
                          std::string &syreq,
                          std::string &asyreq,
                          std::string &async_unsubscribe,
                          unsigned long long* id,
                          std::string &error_msg,
                          long int srv_id) const {
    std::string channel, sid;
    *id = 0;
    bool syreq_without_id, global_u = false;
    std::string m;
    nlohmann::json arr;
    error_msg.clear(); asyreq.clear(); syreq.clear(); async_unsubscribe.clear();
    try {
        nlohmann::json o = nlohmann::json::parse(rin);
        nlohmann::json sysrcs, asysrcs, asy_unsub;
        if(o.find("channel") != o.end())
            channel = o["channel"];
        if(o.find("id") != o.end()) {
            sid = o["id"];
            *id = std::stoull(sid); // catch at the end
        }
        // global method "u" unsubscribe with a valid client id
        // curl http://woody.elettra.eu:8001/bu/src-bundle -d $'{"id":"35", "method":"u"}'
        global_u = o.find("method") != o.end() && o["method"] == "u";
        if(global_u && *id > 0)
            async_unsubscribe = nlohmann::json(o).dump();
        else if(!o.is_null()) {
            arr = o["srcs"];
            if(arr.is_array()) {
                for(auto el : arr) {
                    m = std::string(el["method"]);
                    if(m == "s" && *id > 0) {
                        asysrcs.push_back(el);
                        sysrcs.push_back(el); // also in sync
                    }
                    else if(m == "S" && *id > 0)
                        asysrcs.push_back(el);
                    else if(m == "u" && *id > 0)
                        asy_unsub.push_back(el);
                    else if(m == "s" || m == "u") // sync requests must have client id
                        syreq_without_id |= true;
                    else {
                        sysrcs.push_back(el);
                    }
                }
                nlohmann::json syj(o), asyj(o), u_asyj(o);
                if(asysrcs.size() > 0) {
                    asyj["srcs"] = asysrcs;
                    asyj["srv_id"] = std::to_string(srv_id);
                    asyj["method"] = "s";
                    asyreq = asyj.dump();
                }
                if(sysrcs.size() > 0) {
                    syj["srcs"] = sysrcs;
                    syreq = syj.dump();
                }
                if(asy_unsub.size() > 0) {
                    u_asyj["srcs"] = asy_unsub;
                    u_asyj["srv_id"] = "-1";
                    u_asyj["method"] = "u";
                    async_unsubscribe = u_asyj.dump();
                }
            }
        }
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::divide: JSON parse error: %s in \"%s\"", pe.what(), rin.c_str());
        error_msg = std::string(pe.what());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::divide: nlohmann::json type error: %s", te.what());
        error_msg = std::string(te.what());
    } catch (const std::invalid_argument& ia) {
        error_msg = "CaProxyJsonH: string to number conversion error: " + std::string(ia.what());
    }
    //    if(error_msg.length() == 0 && syreq_without_id)
    //        error_msg = "one or more sync requests without client id";
    if(global_u && *id == 0)
        error_msg = "\"u\" requires \"id\"";
    else if(arr.size() == 0 && !global_u)
        error_msg = "no srcs found";
    else if(m.length() == 0 && !global_u)
        error_msg = "no methods specified";
    //    printf("\e[1;33mCaProxyJsonH::divide DIVIDED rin %s errmsg %s\e[0m\n", rin.c_str(), error_msg.c_str());
    return error_msg.length() == 0;
}

/*!
 * \brief merge sync async subscribe and async unsubscribe replies into a single on.
 * \param syr json sync reply from caserver sync (expected input type: array)
 * \param asyr json async reply from caserver async (subscribe)  (expected: array).
 * \param asyur json async reply from caserver async (unsubscribe) (expected: array).
 * \param error_msg reference to a string that will contain an error message
 *
 * \note if either input json is not an array, the element is added to the returned array
 *
 * \return a Json string with all the results merged in one single array
 *
 */
std::string CaProxyJsonH::merge(const std::string_view& syr, const std::string_view &asyr, const std::string_view &asyur, std::string& error_msg) const {
    std::string s;
    size_t i;
    error_msg.clear();
    nlohmann::json out = nlohmann::json::array(), sy, asy, asyu;


    try {
        if(syr.length() > 0)
            sy = nlohmann::json::parse(syr);
        if(asyr.length() > 0)
            asy = nlohmann::json::parse(asyr);
        if(asyur.length() > 0)
            asyu = nlohmann::json::parse(asyur);

        for(i = 0; !sy.is_null() && sy.is_array() && i < sy.size(); i++)
            out.push_back(sy.at(i));
        if(!sy.is_null() && !sy.is_array())
            out.push_back(sy);
        for(i = 0; !asy.is_null() && asy.is_array() && i < asy.size(); i++)
            out.push_back(asy.at(i));
        if(!asy.is_null() && !asy.is_array())
            out.push_back(asy);
        for(i = 0; !asyu.is_null() && asyu.is_array() && i < asy.size(); i++)
            out.push_back(asyu.at(i));
        if(!asyu.is_null() && !asyu.is_array())
            out.push_back(asyu);

    } catch (const nlohmann::detail::parse_error& pe) {
        error_msg = std::string(pe.what() + std::string(": parse error in one of ") + std::string(syr) + ", " + std::string(asyr) + " or " + std::string(asyur) );
    }  catch (const nlohmann::detail::type_error& te) {
        error_msg = std::string(te.what() + std::string(": type error in one of ") + std::string(syr) + ", " + std::string(asyr)  + " or " + std::string(asyur));
    }

    //    printf("CaProxyJsonH::merge MERGING \e[1;33;2m\"%s\" ----  \e[1;33;2m\"%s\" ---  \e[1;34;2m\"%s\"\e[0m ---> \e[1;32mGOT %s\e[0m\n",
    //           syr.c_str(), asyr.c_str(), asyur.c_str(), out.dump().c_str() );
    return out.dump();
}

// parse s json string
// if s contains "id", convert it to unsigned long long and return it
// if s contains "method":"u" set *unsub to true, false otherwise
// errors if: 1. malformed json 2. method:u without a valid id
//
unsigned long long CaProxyJsonH::tokmsg_process(const std::string &s,
                                                bool *unsub,
                                                std::string &error_msg) {
    *unsub = false;
    unsigned long long tok = 0;
    error_msg.clear();
    if(s.length() == 0)
        return 0;
    try {
        nlohmann::json o = nlohmann::json::parse(s);
        if(!o["id"].is_null()) {
            const std::string& id = o["id"];
            if(id.length() > 0) {
                try {
                    tok = std::stoull(id);
                } catch(const std::invalid_argument &ia) {
                    error_msg = std::string("ca-registrar: invalid argument while extracting id: ") + ia.what();
                }
            }
        }
        if(!o["method"].is_null())
            *unsub = o["method"] == std::string("u");
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::tokmsg_process: JSON parse error: %s in \"%s\"", pe.what(), s.c_str());
        error_msg = std::string(pe.what());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::tokmsg_process: nlohmann::json type error: %s", te.what());
        error_msg = std::string(te.what());
    }
    if(*unsub && tok == 0)
        error_msg = "u option with id 0";
    return tok;
}

std::string CaProxyJsonH::make_unsubscribe(const std::vector<unsigned long long>& ids) const {
    nlohmann::json j;
    j["method"] = "u";
    nlohmann::json jids= nlohmann::json::array();
    for(const unsigned long long& id : ids)
        jids.push_back(std::to_string(id));
    j["ids"] = jids;
    return j.dump();
}

std::string CaProxyJsonH::make_up2date(const std::vector<unsigned long long> &ids) const {
    nlohmann::json j;
    j["method"] = "upd";
    std::string zids = CaIDListH().compress(ids); // "zipped" ids
    j["ids"] = zids;
    return j.dump();
}

std::string CaProxyJsonH::make_err(const std::string &msg) const {
    nlohmann::json je;
    je["err"] = true;
    je["msg"] = msg;
    return je.dump();
}

bool CaProxyJsonH::is_err(const std::string &json) const
{
    try {
        nlohmann::json ao = nlohmann::json::parse(json);
        bool err = (ao.contains("err") && ao["err"]) || ( ao.contains("msg") && ao["msg"] != "ok");
        return err;
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::is_err: JSON parse error: %s in \"%s\"", pe.what(), json.c_str());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::is_err: nlohmann::json type error: %s", te.what());
    }
    return true;
}

// return true if one of the following is true
// json contains "msg" == "ok"
// json contains "ok" = true (boolean)
// json contains "err" and "err" is false (boolean)
//
bool CaProxyJsonH::json_reply_ok(const std::string_view &json) const
{
    try {
        bool ok;
        nlohmann::json o;
        nlohmann::json ao = nlohmann::json::parse(json);
        if(ao.is_array() && ao.size() > 0) {
            o = ao.at(0);
        }
        else if(!ao.is_array()) {
            o = ao;
        }
        if(!o["msg"].is_null() && o["msg"].is_string())
            return o["msg"] == "ok";
        if(o.contains("ok") && (ok = o["ok"]))
            return true;
        if(o.contains("err") && (ok = !o["err"]))
            return true;
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::json_reply_ok: JSON parse error: %s in \"%.*s\"", pe.what(), (int)json.length(), json.data());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::json_reply_ok: nlohmann::json type error: %s", te.what());
    }
    return false;
}

bool CaProxyJsonH::is_subscribe_req(const std::string &json) const {
    try {
        nlohmann::json o = nlohmann::json::parse(json);
        if(!o["method"].is_null() && o["method"].is_string())
            return o["method"] == "s" || o["method"] == "S";
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::is_subscribe_req: JSON parse error: %s in \"%s\"", pe.what(), json.c_str());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::is_subscribe_req: nlohmann::json type error: %s", te.what());
    }
    return false;
}

//bool CaProxyJsonH::insert_srvid_evtyp(const std::string &in,
//                                      long srv_id,
//                                      std::string& out,
//                                      CaProxyEvent::CaProxyEventType ty) const {
//    try {
//        nlohmann::json o = nlohmann::json::parse(in);
//        o["srv_id"] = std::to_string(srv_id); // the other end uses cajson-src-bundle extract (string based)
//        // now set the method globally for ca-supervisor (CaSupDbActivity::execute)
//        if(ty == CaProxyEvent::CaPESub) o["method"] = "s";
//        else if(ty == CaProxyEvent::CaPEUnsub) o["method"] = "u";
//        else if(ty == CaProxyEvent::CaPEUp2Date) o ["method"] = "upd";
//        out = o.dump();
//        return true;
//    }  catch (const nlohmann::detail::parse_error& pe) {
//        perr("CaProxyJsonH::insert_srvid_evtyp: JSON parse error: %s in \"%s\"", pe.what(), in.c_str());
//    }  catch (const nlohmann::detail::type_error& te) {
//        perr("CaProxyJsonH::insert_srvid_evtyp: nlohmann::json type error: %s", te.what());
//    }
//    return false;
//}

/// TEST
///
bool CaProxyJsonH::srcs_match(const std::string &jout, const std::string& jin, int* nsrcout, int *nsrcin) const {
    std::set<std::string> srco, srcin;

    try {
        nlohmann::json o = nlohmann::json::parse(jout); // request to caserver
        if(!o.is_null()) {
            nlohmann::json arr = o["srcs"];
            if(arr.is_array()) {
                for(const auto& el : arr) {
                    std::vector<std::string> keys { "src" };
                    for(const std::string& key : keys) {
                        if(el.contains(key)) {
                            std::string src = el["src"];
                            srco.insert(src);
                        }
                    }
                }
            }
            else
                perr("CaProxyJsonH::srcs_match: JSON srcs to caserver is not an array: %s\n", jout.c_str());

        } // arr.is_array
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::srcs_match: JSON parse error: %s in \"%s\"", pe.what(), jout.c_str());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::srcs_match: nlohmann::json type error: %s", te.what());
    }
    *nsrcout = srco.size();

    // 2.
    try {
        nlohmann::json o = nlohmann::json::parse(jin); // reply from caserver
        if(!o.is_null() && o.is_array()) {
            for(const auto& el : o) {
                std::vector<std::string> keys { "src" };
                for(const std::string& key : keys) {
                    if(el.contains(key)) {
                        std::string src = el["src"];
                        srcin.insert(src);
                    }
                }

            }
        }
        else
            perr("CaProxyJsonH::srcs_match: JSON from caserver is not an array: %s\n", jin.c_str());
    }  catch (const nlohmann::detail::parse_error& pe) {
        perr("CaProxyJsonH::srcs_match: JSON parse error in JSON input: %s in \"%s\"", pe.what(), jin.c_str());
    }  catch (const nlohmann::detail::type_error& te) {
        perr("CaProxyJsonH::srcs_match: nlohmann::json type error: %s", te.what());
    }
    *nsrcin = srcin.size();

    return srco == srcin;
}
///

