#include "caproxythreadpool.h"
#include <vector>
#include <list>
#include <cumbia.h>
#include <cuthreadfactoryimpl.h>
#include <cuthreadseventbridge.h>
#include <cuservices.h>
#include <cuserviceprovider.h>
#include <sys/time.h>  // gettimeofday for srand()
#include <stdlib.h>  // rand
#include <math.h> // round
#include <thread> // for hardware_concurrency
#include <set>  // stores the set of open sockets
#include <ca-receiver-a.h>

class CaProxyThreadPoolPrivate {
public:
    CaProxyThreadPoolPrivate(Cumbia* cu,
                             CaReceiver_A *rece,
                             CuThreadListener *tl,
                             const CuData& options,
                             const std::string& servaddr,
                             const unsigned short servport,
                             int poolsiz) :
        siz(poolsiz), srv_id(-1), cumbia(cu), rec(rece), thl(tl), server_addr(servaddr), server_port(servport), opts(options) { }
    int siz;
    int thidx;
    long srv_id; // our caserver-async id from cadb service table
    std::vector<CaProxy *> pr_activities;
    Cumbia* cumbia;
    CaReceiver_A *rec;
    CuThreadListener *thl;
    std::string server_addr;
    unsigned short server_port;
    const CuData opts;
    std::set<int> open_sockset;
};

CaProxyThreadPool::CaProxyThreadPool( Cumbia* cumbia,
                                      CaReceiver_A *rec,
                                      CuThreadListener *thl,
                                      const CuData& opts,
                                      const std::string& servaddr,
                                      const unsigned short servport,
                                      int poolsiz) {
    d = new CaProxyThreadPoolPrivate(cumbia, rec, thl, opts, servaddr, servport,
                                     poolsiz > 0 ? poolsiz : std::thread::hardware_concurrency());
}

CaProxyThreadPool::~CaProxyThreadPool() {
    printf("CaProxyThreadPool::~CaProxyThreadPool %p deleting %ld activities\n", this, d->pr_activities.size());
    stop();
    delete d;
}

// httphdr: original http header forwarded as transparently as possible
// clen_orig: if message splitting leaves content length unchanged, skip
void CaProxyThreadPool::do_proxy(int fd,
                                 CaProxyEvent::CaPHttpMsgType msgt,
                                 const std::string &syr,
                                 const std::string &s_asyr,
                                 const std::string &u_asyr,
                                 const std::string &httphdr,
                                 unsigned long clen_orig) const {
    if(d->open_sockset.count(fd) > 0) {
        CaProxyEvent *e = new CaProxyEvent(fd, msgt, syr, s_asyr, u_asyr, httphdr, clen_orig);
        CaProxy *t = m_pick_proxy();
        t->proxy_event(e);
    }
}

/*!
 * \brief send a message to the provided file descriptor
 * \param fd the file descriptor
 * \param msg the message
 */
void CaProxyThreadPool::msg_send_cli(int fd, CaProxyEvent::CaPHttpMsgType msgt, const std::string &payload)  const{
    if(d->open_sockset.count(fd) > 0) {
        CaProxyEvent *e = new CaProxyEvent(fd, msgt, payload, CaProxyRoute(CaProxyRoute::ToClient), CaProxyEvent::CaPEUndef);
        CaProxy *t = m_pick_proxy();
        printf("CaProxyThreadPool.msg_send_cli picked proxy %p for fd %d msg %s\n", t, fd, payload.c_str());
        t->proxy_event(e);
    }
    else
        printf("\e[1;31mCaProxyThreadPool::msg_send_cli: not sending msg \"%s\" BECAUSE SOCKET %d IS CLOSED\e[0m\n", payload.c_str(), fd);
}

/*!
 * \brief insert or remove a socket file descriptor in the set of open sockets
 * \param fd file descriptor
 * \param open true: socket is open, false socket is close
 */
void CaProxyThreadPool::fd_state_change(int fd, bool open) const {
//    printf("\e[0;33m [0x%lx] CaProxyThreadPool.fd_state_change this %p fd %d open ? %s\e[0m\n", pthread_self(), this, fd, open ? "Y" : "N");
    if(open) d->open_sockset.insert(fd);
    else d->open_sockset.erase(fd);
}

void CaProxyThreadPool::set_srv_id(long srv_id) {
    d->srv_id = srv_id;
    for(size_t i = 0; i < d->pr_activities.size(); i++)
        d->pr_activities[i]->set_srv_id(srv_id);
}

/*!
 * \brief send a message to a caserver
 * \param pr route
 * \param msg message
 *
 * called by caregistrarlistener. Proxy route ToAsync, to send IDs of currently registered apps
 */
void CaProxyThreadPool::msg_send_caserver(const CaProxyRoute& pr,
                                          CaProxyEvent::CaPHttpMsgType msgt,
                                          const std::string &http_payl,
                                          CaProxyEvent::CaProxyEventType ety) const {
    CaProxyEvent *e = new CaProxyEvent(msgt, http_payl, pr, ety);
    CaProxy *t = m_pick_proxy();
    t->proxy_event(e);
}

std::string CaProxyThreadPool::server_addr() const {
    return d->server_addr;
}

unsigned short CaProxyThreadPool::server_port() const {
    return d->server_port;
}

long CaProxyThreadPool::srv_id() const   {
    return d->srv_id;
}

CaProxy *CaProxyThreadPool::m_pick_proxy() const {
    CaProxy *t = nullptr;
    size_t i = 0;
    for(i = 0; i < d->pr_activities.size() && t == nullptr; i++) {
        if(!d->pr_activities[i]->is_busy()) {
            t = d->pr_activities[i];
        }
    }
    if(!t && d->pr_activities.size() <= static_cast<size_t>(d->siz)) {
        const size_t& idx = d->pr_activities.size();
        t = new CaProxy(d->opts, d->server_addr, d->server_port, d->pr_activities.size(), d->srv_id, CuData("proxy", idx), d->rec);
        t->setFlag(CuActivity::CuADeleteOnExit, true);
        printf("\e[1;36m # \e[0m CaProxyThreadPool::m_pick_proxy creating new activity \e[1;31mtok idx %ld\e[0m cuz ALL %ld busy\n",
               idx, idx);
        d->pr_activities.push_back(t);
        d->cumbia->registerActivity(t, d->thl, "proxy" + std::to_string(idx), CuThreadFactoryImpl(), CuThreadsEventBridgeFactory());
    }
    else if(!t) { // already created d->siz activities: reuse one randomly
        struct timeval tv;
        gettimeofday(&tv, nullptr);
        srand(tv.tv_usec);
        double r = rand() * (d->siz - 1.0f) / double(RAND_MAX);
        int idx = round(r);
        printf("CaProxyThreadPool::m_pick_proxy: idx %d activities size %ld d->siz %d rand %f\n", idx, d->pr_activities.size(), d->siz, r);
        t = d->pr_activities[idx];
        printf("\e[1;35m # \e[0mCaProxyThreadPool::m_pick_proxy  pool is max size: reusing activity %d randomly picked. is busy? %s\n",
               idx, t->is_busy() ? "YES" : "NO");
    }
    return t;
}

void CaProxyThreadPool::stop() {
    for(CaProxy *t : d->pr_activities) {
        t->stop();
        d->cumbia->unregisterActivity(t);
    }
    d->pr_activities.clear();
}

