#include "caunsubmulticast.h"
#include <cudata.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <nlohmann/json.hpp>
#include <cahttpmsg_u.h>
#include <caproxyjson_h.h>
#include <cahttpmsg.h>
#include <string_view>

#define _BUFSIZ 512

class CaUnsubThreadPrivate {
public:
    CaUnsubThreadPrivate(const std::string &host, const std::string & port)
        : m_host(host), m_port(port) {}
    const std::string m_host, m_port;
    std::string errmsg;
    std::thread *m_thread;
    std::string buf, reply;
    CaHttpMsgU utils;
};

CaUnsubThread::CaUnsubThread(const std::string &host, const std::string & port) {
    d = new CaUnsubThreadPrivate(host, port);
    d->m_thread = nullptr;
}

CaUnsubThread::~CaUnsubThread() {
    if(d->m_thread && d->m_thread->joinable())
        d->m_thread->join();
    delete d;
}

void CaUnsubThread::setBuf(const std::string &bu) {
    d->buf = bu;
}

std::string CaUnsubThread::reply() const {
    return d->reply;
}

void CaUnsubThread::start() {
    if(d->m_thread && d->m_thread->joinable()) {
        printf("\e[1;31mca-unsub-thread.start: waiting for older thread to join...\e[0m\n");
        d->m_thread->join();
    }
    if(d->m_thread)
        delete d->m_thread;
    d->m_thread = new std::thread(&CaUnsubThread::run, this);
}

std::string CaUnsubThread::host() const {
    return d->m_host;
}

std::string CaUnsubThread::port() const {
    return d->m_port;
}

std::string CaUnsubThread::errmsg() const {
    return d->errmsg;
}

void CaUnsubThread::wait() {
    if(d->m_thread && d->m_thread->joinable())
        d->m_thread->join();
}

void CaUnsubThread::run() {
    int nread = 0, nwrit = 0;
    int fd, retry = 0;
    errno = 0;
    d->errmsg.clear();
    fd = m_connect(d->m_host.c_str(), d->m_port.c_str());
    while(fd > 0 && retry++ <= 2 && nread <= 0 && d->buf.length() > 0) {
        std::string addr;
        // even if ssock is > 0 because an earlier connection attempt succeeded, the corresponding socket may have
        // gone down and up again. If write fails on ssock, try to reopen it
        nwrit = m_sock_write(fd, d->buf.c_str());
        // caserver will reply something that must be sent back on fd
        if(fd > -1 && nwrit > 0) {
            nread = m_sock_read(fd, d->reply);
        }
        // otherwise nread remains negative
        if(nread <= 0) { // try reconnect
            close(fd);
            printf("[0x%lx] CaUnsubThread::run read failed (nread=%d) on fd %d. \e[1;35mCLOSED FD %d\e[0m: reconnecting\n",
                   pthread_self(), nread, fd, fd);
            fd = m_connect(d->m_host.c_str(), d->m_port.c_str());
            printf("[0x%lx] CaUnsubThread::run: after read failed new \e[1;32mfd %d\e[0m\n", pthread_self(), fd);
        }
    }
    if(fd > -1)
        printf("[0x%lx] CaUnsubThread.run: %s:%s exiting: bytes written %d read back %d\n", pthread_self(), d->m_host.c_str(), d->m_port.c_str(), nwrit, nread);
    if(fd > -1 && close(fd) < 0)
        d->errmsg = "ca-unsub-thread.run: error closing socket: " + std::string(strerror(errno));
    else if(fd > -1)
        printf("[0x%lx] CaUnsubThread::run: \e[1;32msuccessfully\e[1;35m closed fd %d\e[0m\n", pthread_self(), fd);
}

int CaUnsubThread::m_sock_read(int sofd, std::string &in) {
    d->errmsg.clear();
    in.clear();
    char rbuf[_BUFSIZ];

    int nread, totread = 0;
    bool complete = false;
    CaHttpMsg httpa;
    while(!complete && (nread = read(sofd, rbuf, _BUFSIZ -1)) > 0) {
        rbuf[nread] = '\0';
        complete = httpa.add(rbuf);
        totread += nread;
        if(httpa.state == CaHttpMsg::Error)
            break;
    }
    if(nread < 0)
        d->errmsg = "CaUnsubThread::m_sock_read: socket error: " + std::string(strerror(errno));
    if(!complete || httpa.state == CaHttpMsg::Error)
        d->errmsg = "CaUnsubThread::m_sock_read: invalid HTTP message: \"" + httpa.buf + "\"";
    else
        in = std::move(httpa.buf);

    printf("[0x%lx] \e[0;32mR \e[0;33mUNSUB THREAD <--- async \e[1;36m%s:%s\e[0m %d bytes read from %d\e[0m\n", pthread_self(), d->m_host.c_str(), d->m_port.c_str(), nread, sofd);

    return nread >= 0 ? totread : nread;
}

int CaUnsubThread::m_sock_write(int sofd, const char *buf) {
    d->errmsg.clear();
    int totclibw = 0, clibw = 0, wlen = strlen(buf) + 1; // include \0
    // write rbuf back to client fd
    while(totclibw < wlen && clibw >= 0) {
        clibw = send(sofd, buf + totclibw, wlen - totclibw < _BUFSIZ ? wlen - totclibw : _BUFSIZ, MSG_NOSIGNAL);
        if(clibw < 0)
            d->errmsg = "ca-unsub-thread.m_sock_write: write failed : " + std::string(strerror(errno));
        totclibw += clibw;
    }
    return clibw >=0 ? totclibw : clibw;
}

int CaUnsubThread::m_connect(const char *addr, const char *port) {
    int casfd = -1;
    int  s;
    struct addrinfo hints;
    struct addrinfo *result = nullptr, *rp = nullptr;
    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
    hints.ai_socktype = SOCK_STREAM; /* TCP socket */
    hints.ai_flags = 0;
    hints.ai_protocol = 0;          /* Any protocol */
    hints.ai_flags |= AI_NUMERICSERV;
    d->errmsg.clear();
    s = getaddrinfo(addr, port, &hints, &result);
    if (s != 0) {
        d->errmsg = "caproxy.m_connect: getaddrinfo failed: " + std::string(gai_strerror(s));
    }
    else {
        for (rp = result; rp != NULL; rp = rp->ai_next) {
            casfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
            if (casfd == -1)
                continue;
            if (connect(casfd, rp->ai_addr, rp->ai_addrlen) != -1) {
                break;                  /* Success */
            }
            close(casfd);
            casfd = -1;
        }

        if(rp == nullptr) {
            d->errmsg = std::string(addr) + ":" + std::string(port) + ": " + std::string(strerror(errno));
            casfd = -1;
        }
        freeaddrinfo(result);           /* No longer needed */
    }
    return casfd;
}

class CaUnsubMulticastPrivate {
public:
    CaUnsubMulticastPrivate(const CuData& o) : options(o) {}
    const CuData options;
    std::vector<CaUnsubThread *> uthreads;
    std::vector<std::string> errors;
};

CaUnsubMulticast::CaUnsubMulticast(const CuData& opts)
{
    d = new CaUnsubMulticastPrivate(opts);
    std::vector<std::string> hosts = opts["unsubscribe:async-server-address"].toStringVector();
    std::vector<std::string> ports = opts["unsubscribe:async-server-port"].toStringVector();
    for(size_t i = 0; i < hosts.size() && hosts.size() == ports.size(); i++) {
        CaUnsubThread *t = new CaUnsubThread(hosts[i], ports[i]);
        d->uthreads.push_back(t);
    }
}

CaUnsubMulticast::~CaUnsubMulticast() {
    for(CaUnsubThread *t : d->uthreads) {
        t->wait();
        delete t;
    }
    delete d;
}

// send in to all configured caservers for unsubscription
// reply is a HTTP message. The payload "replies" json key
// shall contain the replies from all caservers merged together
// A top level "msg" shall be set to ok if *at least one caserver
// replies successfully*, an "err" field shall be set to true
// otherwise (no caserver replies successfully)
//
// Return value:
// - 200: at least one caserver replies OK
// - 500: all caservers are unreachable (network error)
// - 400: at least one caserver replied, but no reply
//        is OK.
int CaUnsubMulticast::send_multi(const std::string &in, std::string &reply) {
    int code = -1;
    bool ok = false, proc_ok = false;
    reply.clear();
    d->errors.clear();
    std::string payload;
    std::string multirep;
    CaProxyJsonH jh;
    printf("1. [0x%lx] CaUnsubMulticast::send_multi: starting all %ld unsub threads: sending message\n", pthread_self(), d->uthreads.size());
    for(CaUnsubThread * t : d->uthreads) {
        t->setBuf(in);
        t->start();
    }
    // wait for all finished
    printf("2. [0x%lx]  CaUnsubMulticast::send_multi: waiting for all %ld unsub threads\n", pthread_self(), d->uthreads.size());
    for(CaUnsubThread *t : d->uthreads) {
        t->wait();
    }
    for(CaUnsubThread *t : d->uthreads) {
        // one caserver up is enough from the client POV
        ok |= t->errmsg().length() == 0; // t sets error message only on __network error__
        printf("[0x%lx] CaUnsubMulticast::send_multi: \t * errmsg %s\n", pthread_self(), t->errmsg().c_str());
        if(t->errmsg().length() > 0) {
            json_reply_merge(jh.make_err(t->errmsg()), payload, t->host(), t->port(), ok);
        }
        else if(t->reply().length() > 0) {
            std::string_view pl = CaHttpMsg(t->reply()).payload();
            // one proc_ok is enough from the client POV
            proc_ok |= jh.json_reply_ok(pl); // error processing (400 type), not a net err (500)
            // proc_ok determines err=true in top level json data
            json_reply_merge(pl, payload, t->host(), t->port(), proc_ok);
        }
    }
    // forge an HTTP message according to conn_ok and the json merged payload
    CaHttpMsg m;
    // if !conn_ok --> 500 else if proc_ok then 200 else 400
    code = ok ? (proc_ok ? 200 : 400) : 500;
    m.make_reply(code, payload, false); // false: do not set Connection: close in hdr
    reply = std::move(m.buf);
    printf("4. [0x%lx] CaUnsubMulticast::send_multi sending back %s: ok ? %s\e[0m\n",pthread_self(), reply.c_str(), ok  ? "YES" : "NO");
    return code;
}

bool CaUnsubMulticast::json_reply_merge(const std::string_view &jin,
                                        std::string& jout,
                                        const std::string& h,
                                        const std::string& p,
                                        bool ok) const {
    CaProxyJsonH jh;
    std::string err;
    nlohmann::json o, oin, arr;
    if(jout.length() > 0) {
        try {
            o = nlohmann::json::parse(jout);
        } catch (const nlohmann::detail::parse_error& pe) {
            err = "CaUnsubMulticast::json_reply_merge: [1] JSON parse error: " + std::string(pe.what()) + ": " + jout;
        }  catch (const nlohmann::detail::type_error& te) {
            err = "CaUnsubMulticast::json_reply_merge: [1] JSON type error: " + std::string(te.what()) + ": " + jout;
        }
    }
    try{
        oin = nlohmann::json::parse(jin);
    } catch (const nlohmann::detail::parse_error& pe) {
        err = "CaUnsubMulticast::json_reply_merge: [2] JSON parse error: " + std::string(pe.what()) + ": " + std::string(jin);
    }  catch (const nlohmann::detail::type_error& te) {
        err = "CaUnsubMulticast::json_reply_merge: [2] JSON type error: " + std::string(te.what()) + ": " + std::string(jin);
    }
    if(o.is_null()) {
        o = nlohmann::json();
        arr = nlohmann::json::array();
    }
    else if(o.contains("replies"))
        arr = o["replies"];

    if(err.length() > 0 || oin.is_null()) {
        oin = nlohmann::json();
        oin["msg"] = err;
        oin["host"] = h;
        oin["port"] = p;
        arr.push_back(nlohmann::json(err));
    }
    else if(arr.size() == 0 || arr[arr.size() - 1] != oin)
        ;
    arr.push_back(oin);
    int nasy = 0;
    if(o.contains("async-srv"))
        nasy = o["async-srv"];
    nasy++;
    o["async-srv"] = nasy;
    o["replies"] = arr;
    ok ? o["msg"] = "ok" : o["err"] = true;
    jout = o.dump();
    return err.length() == 0;
}
