#include <stdio.h>
#include <stdlib.h>
#include <cumacros.h>
#include <cumbia.h>
#include <cueventloop.h>
#include <signal.h>
#include <cuserviceprovider.h>
#include <cuthreadfactoryimpl.h>
#include <cuthreadseventbridge.h>
#include <pwd.h>
#include <unistd.h>

#include "ca-proxy-process.h"
#include "ca-registrar.h"
#include "caregistrarlistener.h"
#include "caproxythreadpool.h"
#include "caproxylistener.h"
#include "config.h"

#include <culog.h>

#include <ca-receiver-a.h>
#include <calog.h>
#include <cadbhandle.h>
#include <ca-opt-parser.h>

CuEventLoopService *loo_s = nullptr;

void on_int(int signo) {
    if(signo == SIGINT || signo == SIGTERM) {
        if(loo_s)
            loo_s->exit();
    }
}

int drop_privileges(const char* user, char* msg) {
    struct passwd *p = getpwnam(user);
    if(p == nullptr) {
        snprintf(msg, 512, "failed to find UID for user \"%s\": %s", user, strerror(errno));
        return 1;
    }
    if(setgid(p->pw_gid) != 0) {
        snprintf(msg, 512, "unable to drop group privileges to GID %d: %s", p->pw_gid, strerror(errno));
        return 1;
    }
    if(setuid(p->pw_uid) != 0) {
        snprintf(msg, 512, "unable to drop user privileges to UID %d: %s", p->pw_uid, strerror(errno));
        return 1;
    }
    if (setuid(0) != -1)  {
        snprintf(msg, 512, "setuid back to zero succeeded, quitting as this is a security risk");
        return 1;
    }
    msg[0] = '\0';
    return 0;
}

int main(int argc, char *argv[]) {
    CaOptParser options;
    CuData opts = options.get_options(argc, argv, "ca-proxy");
    if(options.errors.size() > 0) {
        perr("main.cpp: error in either command line args or options: \n");
        for(const std::string& e : options.errors)
            perr("- %s\n", e.c_str());
        return EXIT_FAILURE;
    }

    Cumbia *cumbia = new Cumbia();
    // logging - 1. level
    CaLogFactory logfa;
    CuLog *log = nullptr;
    CuLogImplI *log_i = nullptr;
    CuLog::Level logle = static_cast<CuLog::Level>(opts["log_level"].toInt());
    CaProxyProcess *proxy_proc = nullptr;
    CaRegistrar *registrar = nullptr;
    CaRegistrarListener *regilis = nullptr;
    CaProxyListener *proxylis = nullptr;
    CaProxyThreadPool *proxythpoo = nullptr;
    // logging - 2. where
    // drop privileges after this so that the log file can be created anywhere
    // CaFLog will change the file ownership to the unprivileged user after creation
    if(opts.has("log_where", "syslog")) log_i = logfa.create(CaLogFactory::Syslog, logle);
    else if(opts["log_where"].toString() != "console") log_i = logfa.create(CaLogFactory::LogFile, logle, opts["log_where"].toString().c_str(), opts.containsKey("user") ? vtoc2(opts, "user") : nullptr);
    else log_i = logfa.create(CaLogFactory::LogConsole, logle);
    if(log_i) log = new CuLog(log_i);
    // end logging setup

    log->write("main.cpp", "enter: " + std::string(argv[0]) + " version " + std::string(CAPROXY_VERSION), CuLog::LevelAll);

    if(getuid() == 0 && !opts.containsKey("user")) {
        log_i->write("main.cpp", "cannot run as root and \"-s username\" command line argument missing", CuLog::LevelError);
        return EXIT_FAILURE;
    }
    else if(getuid() == 0 && opts.containsKey("user")) {
        char msg[512];
        if(drop_privileges(vtoc2(opts, "user"), msg) != 0) {
            log_i->write("main.cpp", msg, CuLog::LevelError);
            return EXIT_FAILURE;
        }
    }

    bool oerr = false;
    std::vector<std::string> needo { "sync-server-address", "sync-server-port",
                                     "async-server-address", "async-server-port",
                                     "unsubscribe:async-server-address", "unsubscribe:async-server-port" };
    for(const std::string& o : needo) {
        if(!opts.containsKey(o)) {
            oerr |= true;
            perr("main.cpp: missing option \"%s\" in file %s {%s}", o.c_str(), opts.s("conf_file").c_str(), datos(opts));
        }
    }
    if(oerr)
        return EXIT_FAILURE;

    signal(SIGINT, on_int);
    signal(SIGTERM, on_int);

    // The event loop is a single object shared across all Cumbia instances
    loo_s = new CuEventLoopService();
    // register the service with the shared option set to true
    cumbia->getServiceProvider()->registerSharedService(CuServices::EventLoop, loo_s);
    CuThreadFactoryImpl tfi;
    CuThreadsEventBridgeFactory ebf;

    CaReceiver_A *server_a = nullptr;
    CuData srvt("activity", "ca-registrar-srv");
    if(opts.containsKey("bind-address"))
        srvt.set("host", opts.s("bind-address"));
    if(opts.containsKey("port")) {
        srvt.set("port", opts.I("port"));
        server_a = new CaReceiver_A(srvt);
        regilis = new CaRegistrarListener();
        proxylis = new CaProxyListener();
        registrar = new CaRegistrar(opts);
        proxythpoo = new CaProxyThreadPool(cumbia, server_a, proxylis, opts, server_a->address(), server_a->port(), -1);
        proxy_proc = new CaProxyProcess(cumbia, opts, log, server_a->address(), server_a->port(), registrar, proxythpoo);
        server_a->setLog(*log);
        cumbia->registerActivity(registrar, regilis, "ca-registrar", tfi, ebf);
        cumbia->registerActivity(server_a, proxy_proc, "ca-proxy-srv", tfi, ebf);
        regilis->init(proxythpoo, log_i);
        proxylis->init(proxythpoo, log_i);
    }
    if(!opts.containsKey("supervisor-server-address") || !opts.containsKey("supervisor-server-port"))
        printf("\e[1;33m # \e[0mmain.cpp: supervisor-server-address and supervisor-server-port options missing: will not use ca-supervisor\n");

    printf("\e[1;32m # \e[0mmain.cpp: \e[1;32mca-proxy version \e[2;32m%s\e[0m started\n", CAPROXY_VERSION);
    loo_s->exec();
    printf("leaving loop\n");
    if(proxy_proc) {
        delete proxy_proc;
    }
    if(server_a) {
        server_a->stop();
        cumbia->unregisterActivity(server_a);
    }
    if(registrar) {
        registrar->stop();
        cumbia->unregisterActivity(registrar);
    }
    if(proxythpoo) {
        proxythpoo->stop();
        delete proxythpoo;
    }

    delete cumbia;
    if(log_i) {
        log_i->write("ca-proxy", "exiting", CuLog::LevelAll);
        delete log_i;
        delete log;
    }

    if(proxylis) delete proxylis;
    if(regilis) delete regilis;
}
